#ifndef PANGULU_CHECK_H
#define PANGULU_CHECK_H
#ifdef __cplusplus
extern "C"
{
#endif
#include "pangulu_common.h"
    void pangulu_pangulu_Smatrix_U_multiple_pangulu_vector_X(int rank,
                                                             int P,
                                                             int Q,
                                                             int NB,
                                                             int *U_pangulu_Smatrix_rowpointer,
                                                             int *U_pangulu_Smatrix_columnindex,
                                                             pangulu_Smatrix **U_pangulu_Smatrix_value,
                                                             int block_length,
                                                             pangulu_vector *X,
                                                             pangulu_vector *B)
    {
        for (int i = rank / Q; i < block_length; i += P)
        {
            for (int j = U_pangulu_Smatrix_rowpointer[i]; j < U_pangulu_Smatrix_rowpointer[i + 1]; j++)
            {
                int col = U_pangulu_Smatrix_columnindex[j] * NB;
                int row = i * NB;
                pangulu_block_pangulu_Smatrix_multiply_block_pangulu_vector(U_pangulu_Smatrix_value[j],
                                                                            X->value + col,
                                                                            B->value + row);
            }
        }
    }

    void pangulu_pangulu_Smatrix_L_multiple_pangulu_vector_X(int rank,
                                                             int P,
                                                             int Q,
                                                             int NB,
                                                             int *L_pangulu_Smatrix_columnpointer,
                                                             int *L_pangulu_Smatrix_rowindex,
                                                             pangulu_Smatrix **L_pangulu_Smatrix_value,
                                                             int block_length,
                                                             pangulu_vector *X,
                                                             pangulu_vector *B)
    {
        for (int i = rank % Q; i < block_length; i += Q)
        {
            for (int j = L_pangulu_Smatrix_columnpointer[i]; j < L_pangulu_Smatrix_columnpointer[i + 1]; j++)
            {
                int col = i * NB;
                int row = L_pangulu_Smatrix_rowindex[j] * NB;
                pangulu_block_pangulu_Smatrix_multiply_block_pangulu_vector(L_pangulu_Smatrix_value[j],
                                                                            X->value + col,
                                                                            B->value + row);
            }
        }
    }

    void pangulu_gather_pangulu_vector_to_rank_0(int rank,
                                                 pangulu_vector *gather_v,
                                                 int vector_length,
                                                 int P,
                                                 int Q)
    {
        if (rank == 0)
        {
            pangulu_vector *save_vector = (pangulu_vector *)malloc(sizeof(pangulu_vector));
            pangulu_init_pangulu_vector(save_vector, vector_length);
            for (int i = 1; i < P * Q; i++)
            {
                pangulu_recv_pangulu_vector_value(save_vector, i, i, vector_length);
                for (int i = 0; i < vector_length; i++)
                {
                    gather_v->value[i] += save_vector->value[i];
                }
            }
            for (int i = 1; i < P * Q; i++)
            {
                pangulu_send_pangulu_vector_value(gather_v, i, i, vector_length);
            }
            free(save_vector->value);
            free(save_vector);
        }
        else
        {
            pangulu_send_pangulu_vector_value(gather_v, 0, rank, vector_length);
            pangulu_recv_pangulu_vector_value(gather_v, 0, rank, vector_length);
        }
    }

    void pangulu_check_answer(pangulu_vector *X1, pangulu_vector *X2, int n)
    {
        calculate_type error = 0.0;
        for (int i = 0; i < n; i++)
        {
            error += fabs(X1->value[i] - X2->value[i]);
        }
        printf("the error is %f\n", error);
    }

    void pangulu_check(int N,
                       int rank,
                       int P,
                       int Q,
                       int NB,
                       int block_length,
                       pangulu_Smatrix *symbolic_Smatrix,
                       int *L_pangulu_Smatrix_columnpointer,
                       int *L_pangulu_Smatrix_rowindex,
                       pangulu_Smatrix **L_pangulu_Smatrix_value,
                       int *U_pangulu_Smatrix_rowpointer,
                       int *U_pangulu_Smatrix_columnindex,
                       pangulu_Smatrix **U_pangulu_Smatrix_value)
    {
        int vector_length = ((N + NB - 1) / NB) * NB;

        pangulu_vector *X = (pangulu_vector *)malloc(sizeof(pangulu_vector));
        pangulu_read_pangulu_vector(X, vector_length);
        pangulu_vector *b1 = (pangulu_vector *)malloc(sizeof(pangulu_vector));
        pangulu_init_pangulu_vector(b1, vector_length);

        pangulu_pangulu_Smatrix_multiple_pangulu_vector(symbolic_Smatrix, X, b1);

        pangulu_vector *b2 = (pangulu_vector *)malloc(sizeof(pangulu_vector));
        pangulu_read_pangulu_vector(b2, vector_length);

        pangulu_vector *b3 = (pangulu_vector *)malloc(sizeof(pangulu_vector));
        pangulu_init_pangulu_vector(b3, vector_length);

        pangulu_vector *b4 = (pangulu_vector *)malloc(sizeof(pangulu_vector));
        pangulu_init_pangulu_vector(b4, vector_length);
        pangulu_pangulu_Smatrix_U_multiple_pangulu_vector_X(rank,
                                                            P,
                                                            Q,
                                                            NB,
                                                            U_pangulu_Smatrix_rowpointer,
                                                            U_pangulu_Smatrix_columnindex,
                                                            U_pangulu_Smatrix_value,
                                                            block_length,
                                                            b2,
                                                            b3);

        pangulu_gather_pangulu_vector_to_rank_0(rank, b3, vector_length, P, Q);
        pangulu_pangulu_Smatrix_L_multiple_pangulu_vector_X(rank,
                                                            P,
                                                            Q,
                                                            NB,
                                                            L_pangulu_Smatrix_columnpointer,
                                                            L_pangulu_Smatrix_rowindex,
                                                            L_pangulu_Smatrix_value,
                                                            block_length,
                                                            b3,
                                                            b4);

        pangulu_gather_pangulu_vector_to_rank_0(rank, b4, vector_length, P, Q);
        if (rank == 0)
        {
            pangulu_check_answer(b1, b4, N);
        }

        X = pangulu_destroy_pangulu_vector(X);
        b1 = pangulu_destroy_pangulu_vector(b1);
        b2 = pangulu_destroy_pangulu_vector(b2);
        b3 = pangulu_destroy_pangulu_vector(b3);
        b4 = pangulu_destroy_pangulu_vector(b4);
    }
#ifdef __cplusplus
}
#endif

#endif