#include "pangulu_cuda.h"
void pangulu_cuda_malloc_double(double **cuda_address,
                                size_t size)
{
    cudaMalloc((void **)(cuda_address),
               size * sizeof(double));
    //printf("%d %x\n",size*sizeof(double),cuda_address);
}

int *pangulu_cuda_free_int(int *cuda_address)
{
    cudaFree(cuda_address);
    return NULL;
    //printf("%d %x\n",size*sizeof(double),cuda_address);
}

double *pangulu_cuda_free_double(double *cuda_address)
{
    cudaFree(cuda_address);
    return NULL;
    //printf("%d %x\n",size*sizeof(double),cuda_address);
}

void pangulu_cuda_memcpy_host_to_device_double(double *cuda_address,
                                               double *cpu_address,
                                               size_t size)
{
    cudaMemcpy(cuda_address,
               cpu_address,
               size * sizeof(double),
               cudaMemcpyHostToDevice);
}

void pangulu_cuda_memcpy_device_to_host_double(double *cpu_address,
                                               double *cuda_address,
                                               size_t size)
{
    cudaMemcpy(cpu_address,
               cuda_address,
               size * sizeof(double),
               cudaMemcpyDeviceToHost);
}
void pangulu_cuda_malloc_int(int **cuda_address,
                             size_t size)
{
    cudaMalloc((void **)(cuda_address),
               size * sizeof(int));
    //printf("%d %x\n",size*sizeof(double),cuda_address);
}
void pangulu_cuda_memcpy_host_to_device_int(int *cuda_address,
                                            int *cpu_address,
                                            size_t size)
{
    cudaMemcpy(cuda_address,
               cpu_address,
               size * sizeof(int),
               cudaMemcpyHostToDevice);
}

void pangulu_cuda_memcpy_device_to_host_int(int *cpu_address,
                                            int *cuda_address,
                                            size_t size)
{
    cudaMemcpy(cpu_address,
               cuda_address,
               size * sizeof(int),
               cudaMemcpyDeviceToHost);
}

void pangulu_cuda_devicesynchronize()
{
    cudaDeviceSynchronize();
}

void pangulu_cuda_getDevicenum(int *gpu_num)
{
    cudaGetDeviceCount(gpu_num);
}

int pangulu_cuda_setDevice(int gpu_num,
                           int rank)
{
    int usr_id = rank % gpu_num;
    cudaSetDevice(usr_id);
    return usr_id;
}

__global__ void WarpLevel_sflu(int n,
                               int *d_nnzU,
                               int *d_cscColPtrA,
                               int *d_cscRowIdxA,
                               double *d_cscValueA,
                               int *d_cscColPtrL,
                               int *d_cscRowIdxL,
                               double *d_cscValueL,
                               int *d_cscColPtrU,
                               int *d_cscRowIdxU,
                               double *d_cscValueU)
{
    int warp_id = (blockDim.x * blockIdx.x + threadIdx.x) / 32;
    int warp_local_id = threadIdx.x / 32;
    int lane_id = threadIdx.x % 32;

    if (warp_id >= n)
        return;

    const int colidx = warp_id;

    int colA = d_cscColPtrA[colidx];
    int colA1 = d_cscColPtrA[colidx + 1];
    int colU = d_cscColPtrU[colidx];
    int colU1 = d_cscColPtrU[colidx + 1];
    int colL = d_cscColPtrL[colidx];
    int colL1 = d_cscColPtrL[colidx + 1];
    int loadlen = colA1 - colA;

    // use shared memory
    if (loadlen <= SM_LEN_WARPLEV)
    {
        __shared__ int s_idxA[WARP_NUM_WARPLU * SM_LEN_WARPLEV];
        __shared__ double s_valA[WARP_NUM_WARPLU * SM_LEN_WARPLEV];

        int *s_idxA_local = &s_idxA[warp_local_id * SM_LEN_WARPLEV];
        double *s_valA_local = &s_valA[warp_local_id * SM_LEN_WARPLEV];

        for (int i = lane_id; i < loadlen; i += WARP_SIZE)
        {
            s_idxA_local[i] = d_cscRowIdxA[colA + i];
            s_valA_local[i] = d_cscValueA[colA + i];
        }

        //step one
        for (int i = colA; i < colA + colU1 - colU - 1; i++)
        {
            const int rowidx = d_cscRowIdxA[i];
            double value3 = s_valA_local[i - colA];
            //busy-wait until nnzU[rowidx]==0
            do
            {
                __threadfence();
            } while (d_nnzU[rowidx] != -1);

            int rowA = d_cscColPtrA[rowidx];
            int rowA1 = d_cscColPtrA[rowidx + 1];
            int rowU = d_cscColPtrU[rowidx];
            int rowU1 = d_cscColPtrU[rowidx + 1];

            for (int j = rowA + rowU1 - rowU + lane_id; j < rowA1; j += WARP_SIZE)
            {
                const int Lrowindex = d_cscRowIdxA[j];
                const int thecolidx = rowidx;
                int flag1 = binarySearch(s_idxA_local, 0, loadlen - 1, Lrowindex);
                int flag2 = binarySearch(d_cscRowIdxA, d_cscColPtrA[thecolidx] + d_cscColPtrU[thecolidx + 1] - d_cscColPtrU[thecolidx], d_cscColPtrA[thecolidx + 1] - 1, Lrowindex);

                s_valA_local[flag1] -= d_cscValueA[flag2] * value3;
            }

            if (!lane_id)
            {
                atomicSub(&d_nnzU[colidx], 1);
            }
        }

        //step two
        int flag5 = colU1 - colU - 1;
        if (s_valA_local[flag5] > ERROR || s_valA_local[flag5] < -ERROR)
        {
        }
        else
        {
            s_valA_local[flag5] = ERROR;
        }
        double value5 = s_valA_local[flag5];

        d_cscValueL[colL] = 1;
        for (int i = colA + colU1 - colU + lane_id; i < colA1; i += WARP_SIZE)
        {
            const int Lrowindex = s_idxA_local[i - colA];
            int flagL = colL + i - (colA + colU1 - colU) + 1;
            s_valA_local[i - colA] = s_valA_local[i - colA] / value5;
            d_cscValueL[flagL] = s_valA_local[i - colA];
        }

        for (int i = lane_id; i < colU1 - colU; i += WARP_SIZE)
        {
            d_cscValueU[colU + i] = s_valA_local[i];
        }

        for (int i = lane_id; i < loadlen; i += WARP_SIZE)
        {
            d_cscValueA[i + colA] = s_valA_local[i];
        }

        if (!lane_id)
        {
            atomicSub(&d_nnzU[colidx], 1);
        }
    }
    // do not use shared memory
    else
    {
        //step one
        for (int i = colA; i < colA + colU1 - colU - 1; i++)
        {
            const int rowidx = d_cscRowIdxA[i];
            double value3 = d_cscValueA[i];
            //busy-wait until nnzU[rowidx]==0
            do
            {
                __threadfence();
            } while (d_nnzU[rowidx] != -1);

            int rowA = d_cscColPtrA[rowidx];
            int rowA1 = d_cscColPtrA[rowidx + 1];
            int rowU = d_cscColPtrU[rowidx];
            int rowU1 = d_cscColPtrU[rowidx + 1];

            for (int j = rowA + rowU1 - rowU + lane_id; j < rowA1; j += WARP_SIZE)
            {
                const int Lrowindex = d_cscRowIdxA[j];
                const int thecolidx = rowidx;
                int flag1 = binarySearch(d_cscRowIdxA, colA, colA1 - 1, Lrowindex);
                int flag2 = binarySearch(d_cscRowIdxA, d_cscColPtrA[thecolidx] + d_cscColPtrU[thecolidx + 1] - d_cscColPtrU[thecolidx], d_cscColPtrA[thecolidx + 1] - 1, Lrowindex);

                d_cscValueA[flag1] -= d_cscValueA[flag2] * value3;
            }

            if (!lane_id)
            {
                atomicSub(&d_nnzU[colidx], 1);
            }
        }

        for (int i = lane_id; i < colU1 - colU; i += WARP_SIZE)
        {
            d_cscValueU[colU + i] = d_cscValueA[colA + i];
        }

        //step two
        int flag5 = colA + colU1 - colU - 1;
        if (d_cscValueA[flag5] > ERROR || d_cscValueA[flag5] < -ERROR)
        {
        }
        else
        {
            d_cscValueA[flag5] = ERROR;
        }

        double value5 = d_cscValueA[flag5];
        d_cscValueL[colL] = 1;
        for (int i = colA + colU1 - colU + lane_id; i < colA1; i += WARP_SIZE)
        {
            const int Lrowindex = d_cscRowIdxA[i];
            int flagL = colL + i - (colA + colU1 - colU) + 1;
            d_cscValueA[i] = d_cscValueA[i] / value5;
            d_cscValueL[flagL] = d_cscValueA[i];
        }

        if (!lane_id)
        {
            atomicSub(&d_nnzU[colidx], 1);
        }
    }
}

__global__ void BlockLevel_sflu_L1(int n,
                                   int *d_nnzU,
                                   int *d_cscColPtrA,
                                   int *d_cscRowIdxA,
                                   double *d_cscValueA,
                                   int *d_cscColPtrL,
                                   int *d_cscRowIdxL,
                                   double *d_cscValueL,
                                   int *d_cscColPtrU,
                                   int *d_cscRowIdxU,
                                   double *d_cscValueU)
{
    const int colidx = blockIdx.x;

    __shared__ int s_nnzU[1];

    int colA = d_cscColPtrA[colidx];
    int colA1 = d_cscColPtrA[colidx + 1];
    int smemlen = 256;
    int len_a = colA1 - colA;

    // use shared memory
    if (len_a <= smemlen)
    {
        __shared__ int s_rowidxA[256];
        __shared__ double s_valueA[256];

        for (int i = threadIdx.x; i < len_a; i += blockDim.x)
        {
            s_rowidxA[i] = d_cscRowIdxA[i + colA];
            s_valueA[i] = d_cscValueA[i + colA];
        }

        if (!threadIdx.x)
        {
            s_nnzU[0] = d_nnzU[colidx];
        }
        __syncthreads();

        int colU = d_cscColPtrU[colidx];
        int colU1 = d_cscColPtrU[colidx + 1];
        int colL = d_cscColPtrL[colidx];
        int colL1 = d_cscColPtrL[colidx + 1];

        //step one
        for (int i = colA; i < colA + colU1 - colU - 1; i++)
        {
            const int rowidx = d_cscRowIdxA[i];
            double value3 = s_valueA[i - colA];

            //busy-wait until nnzU[rowidx]==0
            if (!threadIdx.x)
            {
                do
                {
                    __threadfence_block();
                } while (d_nnzU[rowidx] != -1);
            }
            __syncthreads();

            int rowA = d_cscColPtrA[rowidx];
            int rowA1 = d_cscColPtrA[rowidx + 1];
            int rowU = d_cscColPtrU[rowidx];
            int rowU1 = d_cscColPtrU[rowidx + 1];

            for (int j = rowA + rowU1 - rowU + threadIdx.x; j < rowA1; j += blockDim.x)
            {
                const int Lrowindex = d_cscRowIdxA[j];
                const int thecolidx = rowidx;

                int flag1 = binarySearch(s_rowidxA, 0, len_a - 1, Lrowindex);
                int flag2 = binarySearch(d_cscRowIdxA, d_cscColPtrA[thecolidx] + d_cscColPtrU[thecolidx + 1] - d_cscColPtrU[thecolidx], d_cscColPtrA[thecolidx + 1] - 1, Lrowindex);

                s_valueA[flag1] -= d_cscValueA[flag2] * value3;
            }

            if (!threadIdx.x)
            {
                atomicSub(&s_nnzU[0], 1);
            }
            __syncthreads();
        }

        //step two
        int flag5 = colU1 - colU - 1;
        if (s_valueA[flag5] > ERROR || s_valueA[flag5] < -ERROR)
        {
        }
        else
        {
            s_valueA[flag5] = ERROR;
        }
        double value5 = s_valueA[flag5];

        d_cscValueL[colL] = 1;
        for (int i = colA + colU1 - colU + threadIdx.x; i < colA1; i += blockDim.x)
        {
            const int Lrowindex = s_rowidxA[i - colA];
            int flagL = colL + i - (colA + colU1 - colU) + 1;
            s_valueA[i - colA] = s_valueA[i - colA] / value5;
            d_cscValueL[flagL] = s_valueA[i - colA];
        }
        __syncthreads();

        for (int i = threadIdx.x; i < colU1 - colU; i += blockDim.x)
        {
            d_cscValueU[colU + i] = s_valueA[i];
        }

        for (int i = threadIdx.x; i < len_a; i += blockDim.x)
        {
            d_cscValueA[i + colA] = s_valueA[i];
        }

        if (!threadIdx.x)
        {
            atomicSub(&s_nnzU[0], 1);
            d_nnzU[colidx] = s_nnzU[0];
        }
    }
    //do not use shared memory
    else
    {
        if (!threadIdx.x)
        {
            s_nnzU[0] = d_nnzU[colidx];
        }
        __syncthreads();

        int colU = d_cscColPtrU[colidx];
        int colU1 = d_cscColPtrU[colidx + 1];
        int colL = d_cscColPtrL[colidx];
        int colL1 = d_cscColPtrL[colidx + 1];

        //step one
        for (int i = colA; i < colA + colU1 - colU - 1; i++)
        {
            const int rowidx = d_cscRowIdxA[i];
            double value3 = d_cscValueA[i];
            //busy-wait until nnzU[rowidx]==0
            if (!threadIdx.x)
            {
                do
                {
                    __threadfence_block();
                } while (d_nnzU[rowidx] != -1);
            }
            __syncthreads();

            int rowA = d_cscColPtrA[rowidx];
            int rowA1 = d_cscColPtrA[rowidx + 1];
            int rowU = d_cscColPtrU[rowidx];
            int rowU1 = d_cscColPtrU[rowidx + 1];

            for (int j = rowA + rowU1 - rowU + threadIdx.x; j < rowA1; j += blockDim.x)
            {
                const int Lrowindex = d_cscRowIdxA[j];
                const int thecolidx = rowidx;
                int flag1 = binarySearch(d_cscRowIdxA, colA, colA1 - 1, Lrowindex);
                int flag2 = binarySearch(d_cscRowIdxA, d_cscColPtrA[thecolidx] + d_cscColPtrU[thecolidx + 1] - d_cscColPtrU[thecolidx], d_cscColPtrA[thecolidx + 1] - 1, Lrowindex);

                d_cscValueA[flag1] -= d_cscValueA[flag2] * value3;
            }

            if (!threadIdx.x)
            {
                atomicSub(&s_nnzU[0], 1);
            }
            __syncthreads();
        }

        for (int i = threadIdx.x; i < colU1 - colU; i += blockDim.x)
        {
            d_cscValueU[colU + i] = d_cscValueA[colA + i];
        }

        //step two
        int flag5 = colA + colU1 - colU - 1;
        if (d_cscValueA[flag5] > ERROR || d_cscValueA[flag5] < -ERROR)
        {
        }
        else
        {
            d_cscValueA[flag5] = ERROR;
        }
        double value5 = d_cscValueA[flag5];
        d_cscValueL[colL] = 1;
        
        for (int i = colA + colU1 - colU + threadIdx.x; i < colA1; i += blockDim.x)
        {
            const int Lrowindex = d_cscRowIdxA[i];
            int flagL = colL + i - (colA + colU1 - colU) + 1;
            d_cscValueA[i] = d_cscValueA[i] / value5;
            d_cscValueL[flagL] = d_cscValueA[i];
        }

        if (!threadIdx.x)
        {
            atomicSub(&s_nnzU[0], 1);
            d_nnzU[colidx] = s_nnzU[0];
        }
    }
}

__global__ void BlockLevel_sflu_L2(int n,
                                   int *d_nnzU,
                                   int *d_cscColPtrA,
                                   int *d_cscRowIdxA,
                                   double *d_cscValueA,
                                   int *d_cscColPtrL,
                                   int *d_cscRowIdxL,
                                   double *d_cscValueL,
                                   int *d_cscColPtrU,
                                   int *d_cscRowIdxU,
                                   double *d_cscValueU)
{
    const int colidx = blockIdx.x;

    __shared__ int s_nnzU[1];

    int colA = d_cscColPtrA[colidx];
    int colA1 = d_cscColPtrA[colidx + 1];
    int smemlen = 512;
    int len_a = colA1 - colA;

    // use shared memory
    if (len_a <= smemlen)
    {
        __shared__ int s_rowidxA[512];
        __shared__ double s_valueA[512];

        for (int i = threadIdx.x; i < len_a; i += blockDim.x)
        {
            s_rowidxA[i] = d_cscRowIdxA[i + colA];
            s_valueA[i] = d_cscValueA[i + colA];
        }

        if (!threadIdx.x)
        {
            s_nnzU[0] = d_nnzU[colidx];
        }
        __syncthreads();

        int colU = d_cscColPtrU[colidx];
        int colU1 = d_cscColPtrU[colidx + 1];
        int colL = d_cscColPtrL[colidx];
        int colL1 = d_cscColPtrL[colidx + 1];

        //step one
        for (int i = colA; i < colA + colU1 - colU - 1; i++)
        {
            const int rowidx = d_cscRowIdxA[i];
            double value3 = s_valueA[i - colA];

            //busy-wait until nnzU[rowidx]==0
            if (!threadIdx.x)
            {
                do
                {
                    __threadfence_block();
                } while (d_nnzU[rowidx] != -1);
            }
            __syncthreads();

            int rowA = d_cscColPtrA[rowidx];
            int rowA1 = d_cscColPtrA[rowidx + 1];
            int rowU = d_cscColPtrU[rowidx];
            int rowU1 = d_cscColPtrU[rowidx + 1];

            for (int j = rowA + rowU1 - rowU + threadIdx.x; j < rowA1; j += blockDim.x)
            {
                const int Lrowindex = d_cscRowIdxA[j];
                const int thecolidx = rowidx;

                int flag1 = binarySearch(s_rowidxA, 0, len_a - 1, Lrowindex);
                int flag2 = binarySearch(d_cscRowIdxA, d_cscColPtrA[thecolidx] + d_cscColPtrU[thecolidx + 1] - d_cscColPtrU[thecolidx], d_cscColPtrA[thecolidx + 1] - 1, Lrowindex);

                s_valueA[flag1] -= d_cscValueA[flag2] * value3;
            }

            if (!threadIdx.x)
            {
                atomicSub(&s_nnzU[0], 1);
            }
            __syncthreads();
        }

        //step two
        int flag5 = colU1 - colU - 1;
        if (s_valueA[flag5] > ERROR || s_valueA[flag5] < -ERROR)
        {
        }
        else
        {
            s_valueA[flag5] = ERROR;
        }
        double value5 = s_valueA[flag5];

        d_cscValueL[colL] = 1;
        for (int i = colA + colU1 - colU + threadIdx.x; i < colA1; i += blockDim.x)
        {
            const int Lrowindex = s_rowidxA[i - colA];
            int flagL = colL + i - (colA + colU1 - colU) + 1;
            s_valueA[i - colA] = s_valueA[i - colA] / value5;
            d_cscValueL[flagL] = s_valueA[i - colA];
        }
        __syncthreads();

        for (int i = threadIdx.x; i < colU1 - colU; i += blockDim.x)
        {
            d_cscValueU[colU + i] = s_valueA[i];
        }

        for (int i = threadIdx.x; i < len_a; i += blockDim.x)
        {
            d_cscValueA[i + colA] = s_valueA[i];
        }

        if (!threadIdx.x)
        {
            atomicSub(&s_nnzU[0], 1);
            d_nnzU[colidx] = s_nnzU[0];
        }
    }
    //do not use shared memory
    else
    {
        if (!threadIdx.x)
        {
            s_nnzU[0] = d_nnzU[colidx];
        }
        __syncthreads();

        int colU = d_cscColPtrU[colidx];
        int colU1 = d_cscColPtrU[colidx + 1];
        int colL = d_cscColPtrL[colidx];
        int colL1 = d_cscColPtrL[colidx + 1];

        //step one
        for (int i = colA; i < colA + colU1 - colU - 1; i++)
        {
            const int rowidx = d_cscRowIdxA[i];

            double value3 = d_cscValueA[i];

            //busy-wait until nnzU[rowidx]==0
            if (!threadIdx.x)
            {
                do
                {
                    __threadfence_block();
                } while (d_nnzU[rowidx] != -1);
            }
            __syncthreads();

            int rowA = d_cscColPtrA[rowidx];
            int rowA1 = d_cscColPtrA[rowidx + 1];
            int rowU = d_cscColPtrU[rowidx];
            int rowU1 = d_cscColPtrU[rowidx + 1];

            for (int j = rowA + rowU1 - rowU + threadIdx.x; j < rowA1; j += blockDim.x)
            {
                const int Lrowindex = d_cscRowIdxA[j];
                const int thecolidx = rowidx;

                int flag1 = binarySearch(d_cscRowIdxA, colA, colA1 - 1, Lrowindex);
                int flag2 = binarySearch(d_cscRowIdxA, d_cscColPtrA[thecolidx] + d_cscColPtrU[thecolidx + 1] - d_cscColPtrU[thecolidx], d_cscColPtrA[thecolidx + 1] - 1, Lrowindex);

                d_cscValueA[flag1] -= d_cscValueA[flag2] * value3;
            }

            if (!threadIdx.x)
            {
                atomicSub(&s_nnzU[0], 1);
            }
            __syncthreads();
        }

        for (int i = threadIdx.x; i < colU1 - colU; i += blockDim.x)
        {
            d_cscValueU[colU + i] = d_cscValueA[colA + i];
        }

        //step two
        int flag5 = colA + colU1 - colU - 1;
        if (d_cscValueA[flag5] > ERROR || d_cscValueA[flag5] < -ERROR)
        {
        }
        else
        {
            d_cscValueA[flag5] = ERROR;
        }
        double value5 = d_cscValueA[flag5];

        d_cscValueL[colL] = 1;
        for (int i = colA + colU1 - colU + threadIdx.x; i < colA1; i += blockDim.x)
        {
            const int Lrowindex = d_cscRowIdxA[i];
            int flagL = colL + i - (colA + colU1 - colU) + 1;
            d_cscValueA[i] = d_cscValueA[i] / value5;
            d_cscValueL[flagL] = d_cscValueA[i];
        }

        if (!threadIdx.x)
        {
            atomicSub(&s_nnzU[0], 1);
            d_nnzU[colidx] = s_nnzU[0];
        }
    }
}

__global__ void BlockLevel_sflu_L3(int n,
                                   int *d_nnzU,
                                   int *d_cscColPtrA,
                                   int *d_cscRowIdxA,
                                   double *d_cscValueA,
                                   int *d_cscColPtrL,
                                   int *d_cscRowIdxL,
                                   double *d_cscValueL,
                                   int *d_cscColPtrU,
                                   int *d_cscRowIdxU,
                                   double *d_cscValueU)
{
    const int colidx = blockIdx.x;

    __shared__ int s_nnzU[1];

    int colA = d_cscColPtrA[colidx];
    int colA1 = d_cscColPtrA[colidx + 1];
    int smemlen = 1024;
    int len_a = colA1 - colA;

    // use shared memory
    if (len_a <= smemlen)
    {
        __shared__ int s_rowidxA[1024];
        __shared__ double s_valueA[1024];

        for (int i = threadIdx.x; i < len_a; i += blockDim.x)
        {
            s_rowidxA[i] = d_cscRowIdxA[i + colA];
            s_valueA[i] = d_cscValueA[i + colA];
        }

        if (!threadIdx.x)
        {
            s_nnzU[0] = d_nnzU[colidx];
        }
        __syncthreads();

        int colU = d_cscColPtrU[colidx];
        int colU1 = d_cscColPtrU[colidx + 1];
        int colL = d_cscColPtrL[colidx];
        int colL1 = d_cscColPtrL[colidx + 1];

        //step one
        for (int i = colA; i < colA + colU1 - colU - 1; i++)
        {
            const int rowidx = d_cscRowIdxA[i];
            double value3 = s_valueA[i - colA];

            //busy-wait until nnzU[rowidx]==0
            if (!threadIdx.x)
            {
                do
                {
                    __threadfence_block();
                } while (d_nnzU[rowidx] != -1);
            }
            __syncthreads();

            int rowA = d_cscColPtrA[rowidx];
            int rowA1 = d_cscColPtrA[rowidx + 1];
            int rowU = d_cscColPtrU[rowidx];
            int rowU1 = d_cscColPtrU[rowidx + 1];

            for (int j = rowA + rowU1 - rowU + threadIdx.x; j < rowA1; j += blockDim.x)
            {
                const int Lrowindex = d_cscRowIdxA[j];
                const int thecolidx = rowidx;

                int flag1 = binarySearch(s_rowidxA, 0, len_a - 1, Lrowindex);
                int flag2 = binarySearch(d_cscRowIdxA, d_cscColPtrA[thecolidx] + d_cscColPtrU[thecolidx + 1] - d_cscColPtrU[thecolidx], d_cscColPtrA[thecolidx + 1] - 1, Lrowindex);

                s_valueA[flag1] -= d_cscValueA[flag2] * value3;
            }

            if (!threadIdx.x)
            {
                atomicSub(&s_nnzU[0], 1);
            }
            __syncthreads();
        }

        //step two
        int flag5 = colU1 - colU - 1;
        if (s_valueA[flag5] > ERROR || s_valueA[flag5] < -ERROR)
        {
        }
        else
        {
            s_valueA[flag5] = ERROR;
        }
        double value5 = s_valueA[flag5];

        d_cscValueL[colL] = 1;
        for (int i = colA + colU1 - colU + threadIdx.x; i < colA1; i += blockDim.x)
        {
            const int Lrowindex = s_rowidxA[i - colA];
            int flagL = colL + i - (colA + colU1 - colU) + 1;
            s_valueA[i - colA] = s_valueA[i - colA] / value5;
            d_cscValueL[flagL] = s_valueA[i - colA];
        }
        __syncthreads();

        for (int i = threadIdx.x; i < colU1 - colU; i += blockDim.x)
        {
            d_cscValueU[colU + i] = s_valueA[i];
        }

        for (int i = threadIdx.x; i < len_a; i += blockDim.x)
        {
            d_cscValueA[i + colA] = s_valueA[i];
        }

        if (!threadIdx.x)
        {
            atomicSub(&s_nnzU[0], 1);
            d_nnzU[colidx] = s_nnzU[0];
        }
    }
    //do not use shared memory
    else
    {
        if (!threadIdx.x)
        {
            s_nnzU[0] = d_nnzU[colidx];
        }
        __syncthreads();

        int colU = d_cscColPtrU[colidx];
        int colU1 = d_cscColPtrU[colidx + 1];
        int colL = d_cscColPtrL[colidx];
        int colL1 = d_cscColPtrL[colidx + 1];

        //step one
        for (int i = colA; i < colA + colU1 - colU - 1; i++)
        {
            const int rowidx = d_cscRowIdxA[i];

            double value3 = d_cscValueA[i];

            //busy-wait until nnzU[rowidx]==0
            if (!threadIdx.x)
            {
                do
                {
                    __threadfence_block();
                } while (d_nnzU[rowidx] != -1);
            }
            __syncthreads();

            int rowA = d_cscColPtrA[rowidx];
            int rowA1 = d_cscColPtrA[rowidx + 1];
            int rowU = d_cscColPtrU[rowidx];
            int rowU1 = d_cscColPtrU[rowidx + 1];

            for (int j = rowA + rowU1 - rowU + threadIdx.x; j < rowA1; j += blockDim.x)
            {
                const int Lrowindex = d_cscRowIdxA[j];
                const int thecolidx = rowidx;

                int flag1 = binarySearch(d_cscRowIdxA, colA, colA1 - 1, Lrowindex);
                int flag2 = binarySearch(d_cscRowIdxA, d_cscColPtrA[thecolidx] + d_cscColPtrU[thecolidx + 1] - d_cscColPtrU[thecolidx], d_cscColPtrA[thecolidx + 1] - 1, Lrowindex);

                d_cscValueA[flag1] -= d_cscValueA[flag2] * value3;
            }

            if (!threadIdx.x)
            {
                atomicSub(&s_nnzU[0], 1);
            }
            __syncthreads();
        }

        for (int i = threadIdx.x; i < colU1 - colU; i += blockDim.x)
        {
            d_cscValueU[colU + i] = d_cscValueA[colA + i];
        }

        //step two
        int flag5 = colA + colU1 - colU - 1;
        if (d_cscValueA[flag5] > ERROR || d_cscValueA[flag5] < -ERROR)
        {
        }
        else
        {
            d_cscValueA[flag5] = ERROR;
        }
        double value5 = d_cscValueA[flag5];

        d_cscValueL[colL] = 1;
        for (int i = colA + colU1 - colU + threadIdx.x; i < colA1; i += blockDim.x)
        {
            const int Lrowindex = d_cscRowIdxA[i];
            int flagL = colL + i - (colA + colU1 - colU) + 1;
            d_cscValueA[i] = d_cscValueA[i] / value5;
            d_cscValueL[flagL] = d_cscValueA[i];
        }

        if (!threadIdx.x)
        {
            atomicSub(&s_nnzU[0], 1);
            d_nnzU[colidx] = s_nnzU[0];
        }
    }
}

__global__ void BlockLevel_sflu_L4(int n,
                                   int *d_nnzU,
                                   int *d_cscColPtrA,
                                   int *d_cscRowIdxA,
                                   double *d_cscValueA,
                                   int *d_cscColPtrL,
                                   int *d_cscRowIdxL,
                                   double *d_cscValueL,
                                   int *d_cscColPtrU,
                                   int *d_cscRowIdxU,
                                   double *d_cscValueU)
{
    const int colidx = blockIdx.x;

    __shared__ int s_nnzU[1];

    int colA = d_cscColPtrA[colidx];
    int colA1 = d_cscColPtrA[colidx + 1];
    int smemlen = 2048;
    int len_a = colA1 - colA;

    // use shared memory
    if (len_a <= smemlen)
    {
        __shared__ int s_rowidxA[2048];
        __shared__ double s_valueA[2048];

        for (int i = threadIdx.x; i < len_a; i += blockDim.x)
        {
            s_rowidxA[i] = d_cscRowIdxA[i + colA];
            s_valueA[i] = d_cscValueA[i + colA];
        }

        if (!threadIdx.x)
        {
            s_nnzU[0] = d_nnzU[colidx];
        }
        __syncthreads();

        int colU = d_cscColPtrU[colidx];
        int colU1 = d_cscColPtrU[colidx + 1];
        int colL = d_cscColPtrL[colidx];
        int colL1 = d_cscColPtrL[colidx + 1];

        //step one
        for (int i = colA; i < colA + colU1 - colU - 1; i++)
        {
            const int rowidx = d_cscRowIdxA[i];
            double value3 = s_valueA[i - colA];

            //busy-wait until nnzU[rowidx]==0
            if (!threadIdx.x)
            {
                do
                {
                    __threadfence_block();
                } while (d_nnzU[rowidx] != -1);
            }
            __syncthreads();

            int rowA = d_cscColPtrA[rowidx];
            int rowA1 = d_cscColPtrA[rowidx + 1];
            int rowU = d_cscColPtrU[rowidx];
            int rowU1 = d_cscColPtrU[rowidx + 1];

            for (int j = rowA + rowU1 - rowU + threadIdx.x; j < rowA1; j += blockDim.x)
            {
                const int Lrowindex = d_cscRowIdxA[j];
                const int thecolidx = rowidx;

                int flag1 = binarySearch(s_rowidxA, 0, len_a - 1, Lrowindex);
                int flag2 = binarySearch(d_cscRowIdxA, d_cscColPtrA[thecolidx] + d_cscColPtrU[thecolidx + 1] - d_cscColPtrU[thecolidx], d_cscColPtrA[thecolidx + 1] - 1, Lrowindex);

                s_valueA[flag1] -= d_cscValueA[flag2] * value3;
            }

            if (!threadIdx.x)
            {
                atomicSub(&s_nnzU[0], 1);
            }
            __syncthreads();
        }

        //step two
        int flag5 = colU1 - colU - 1;
        if (s_valueA[flag5] > ERROR || s_valueA[flag5] < -ERROR)
        {
        }
        else
        {
            s_valueA[flag5] = ERROR;
        }
        double value5 = s_valueA[flag5];

        d_cscValueL[colL] = 1;
        for (int i = colA + colU1 - colU + threadIdx.x; i < colA1; i += blockDim.x)
        {
            const int Lrowindex = s_rowidxA[i - colA];
            int flagL = colL + i - (colA + colU1 - colU) + 1;
            s_valueA[i - colA] = s_valueA[i - colA] / value5;
            d_cscValueL[flagL] = s_valueA[i - colA];
        }
        __syncthreads();

        for (int i = threadIdx.x; i < colU1 - colU; i += blockDim.x)
        {
            d_cscValueU[colU + i] = s_valueA[i];
        }

        for (int i = threadIdx.x; i < len_a; i += blockDim.x)
        {
            d_cscValueA[i + colA] = s_valueA[i];
        }

        if (!threadIdx.x)
        {
            atomicSub(&s_nnzU[0], 1);
            d_nnzU[colidx] = s_nnzU[0];
        }
    }
    //do not use shared memory
    else
    {
        if (!threadIdx.x)
        {
            s_nnzU[0] = d_nnzU[colidx];
        }
        __syncthreads();

        int colU = d_cscColPtrU[colidx];
        int colU1 = d_cscColPtrU[colidx + 1];
        int colL = d_cscColPtrL[colidx];
        int colL1 = d_cscColPtrL[colidx + 1];

        //step one
        for (int i = colA; i < colA + colU1 - colU - 1; i++)
        {
            const int rowidx = d_cscRowIdxA[i];

            double value3 = d_cscValueA[i];

            //busy-wait until nnzU[rowidx]==0
            if (!threadIdx.x)
            {
                do
                {
                    __threadfence_block();
                } while (d_nnzU[rowidx] != -1);
            }
            __syncthreads();

            int rowA = d_cscColPtrA[rowidx];
            int rowA1 = d_cscColPtrA[rowidx + 1];
            int rowU = d_cscColPtrU[rowidx];
            int rowU1 = d_cscColPtrU[rowidx + 1];

            for (int j = rowA + rowU1 - rowU + threadIdx.x; j < rowA1; j += blockDim.x)
            {
                const int Lrowindex = d_cscRowIdxA[j];
                const int thecolidx = rowidx;

                int flag1 = binarySearch(d_cscRowIdxA, colA, colA1 - 1, Lrowindex);
                int flag2 = binarySearch(d_cscRowIdxA, d_cscColPtrA[thecolidx] + d_cscColPtrU[thecolidx + 1] - d_cscColPtrU[thecolidx], d_cscColPtrA[thecolidx + 1] - 1, Lrowindex);

                d_cscValueA[flag1] -= d_cscValueA[flag2] * value3;
            }

            if (!threadIdx.x)
            {
                atomicSub(&s_nnzU[0], 1);
            }
            __syncthreads();
        }

        for (int i = threadIdx.x; i < colU1 - colU; i += blockDim.x)
        {
            d_cscValueU[colU + i] = d_cscValueA[colA + i];
        }

        //step two
        int flag5 = colA + colU1 - colU - 1;
        if (d_cscValueA[flag5] > ERROR || d_cscValueA[flag5] < -ERROR)
        {
        }
        else
        {
            d_cscValueA[flag5] = ERROR;
        }
        double value5 = d_cscValueA[flag5];

        d_cscValueL[colL] = 1;
        for (int i = colA + colU1 - colU + threadIdx.x; i < colA1; i += blockDim.x)
        {
            const int Lrowindex = d_cscRowIdxA[i];
            int flagL = colL + i - (colA + colU1 - colU) + 1;
            d_cscValueA[i] = d_cscValueA[i] / value5;
            d_cscValueL[flagL] = d_cscValueA[i];
        }

        if (!threadIdx.x)
        {
            atomicSub(&s_nnzU[0], 1);
            d_nnzU[colidx] = s_nnzU[0];
        }
    }
}

__device__ int binarySearch(int *ridx, int left, int right, int target)
{
    int mid;
    while (left <= right)
    {
        mid = left + (right - left) / 2;
        if (ridx[mid] == target)
        {
            return mid;
        }
        else if (ridx[mid] > target)
        {
            right = mid - 1;
        }
        else
        {
            left = mid + 1;
        }
    }
    return -1;
}

void pangulu_getrf_cuda_kernel(int n,
                               int nnzA,
                               int *d_nnzU,
                               int *A_CUDA_rowpointer,
                               int *A_CUDA_columnindex,
                               double *A_CUDA_value,
                               int *L_CUDA_rowpointer,
                               int *L_CUDA_columnindex,
                               double *L_CUDA_value,
                               int *U_CUDA_rowpointer,
                               int *U_CUDA_columnindex,
                               double *U_CUDA_value)
{
    int nnz_avrg = nnzA / n;

    int num_blocks;
    int num_threads;

    if (nnz_avrg <= 32)
    {
        num_blocks = ceil((double)n / WARP_NUM_WARPLU);
        num_threads = WARP_NUM_WARPLU * WARP_SIZE;
        WarpLevel_sflu<<<num_blocks, num_threads>>>(n, d_nnzU, A_CUDA_rowpointer, A_CUDA_columnindex, A_CUDA_value,
                                                    L_CUDA_rowpointer, L_CUDA_columnindex, L_CUDA_value,
                                                    U_CUDA_rowpointer, U_CUDA_columnindex, U_CUDA_value);
    }
    else if (nnz_avrg > 32 && nnz_avrg <= 96)
    {
        num_blocks = n;
        num_threads = 2 * WARP_SIZE;
        BlockLevel_sflu_L1<<<num_blocks, num_threads>>>(n, d_nnzU, A_CUDA_rowpointer, A_CUDA_columnindex, A_CUDA_value,
                                                        L_CUDA_rowpointer, L_CUDA_columnindex, L_CUDA_value,
                                                        U_CUDA_rowpointer, U_CUDA_columnindex, U_CUDA_value);
    }
    else if (nnz_avrg > 96 && nnz_avrg <= 192)
    {
        num_blocks = n;
        num_threads = 4 * WARP_SIZE;
        BlockLevel_sflu_L2<<<num_blocks, num_threads>>>(n, d_nnzU, A_CUDA_rowpointer, A_CUDA_columnindex, A_CUDA_value,
                                                        L_CUDA_rowpointer, L_CUDA_columnindex, L_CUDA_value,
                                                        U_CUDA_rowpointer, U_CUDA_columnindex, U_CUDA_value);
    }
    else if (nnz_avrg > 192 && nnz_avrg <= 384)
    {
        num_blocks = n;
        num_threads = 8 * WARP_SIZE;
        BlockLevel_sflu_L3<<<num_blocks, num_threads>>>(n, d_nnzU, A_CUDA_rowpointer, A_CUDA_columnindex, A_CUDA_value,
                                                        L_CUDA_rowpointer, L_CUDA_columnindex, L_CUDA_value,
                                                        U_CUDA_rowpointer, U_CUDA_columnindex, U_CUDA_value);
    }
    else
    {
        num_blocks = n;
        num_threads = 16 * WARP_SIZE;
        BlockLevel_sflu_L4<<<num_blocks, num_threads>>>(n, d_nnzU, A_CUDA_rowpointer, A_CUDA_columnindex, A_CUDA_value,
                                                        L_CUDA_rowpointer, L_CUDA_columnindex, L_CUDA_value,
                                                        U_CUDA_rowpointer, U_CUDA_columnindex, U_CUDA_value);
    }
}

__global__ void trans_cuda(int n, double *d_val_csr, int *d_idx, double *d_val_csc)
{

    if (blockDim.x * blockIdx.x + threadIdx.x >= n)
        return;

    int i = blockDim.x * blockIdx.x + threadIdx.x;
    d_val_csc[d_idx[i]] = d_val_csr[i];
}

void pangulu_cuda_transport_kernel(int n, double *d_val_csr, int *d_idx, double *d_val_csc)
{

    int num_threads = WARP_SIZE * 2;
    int num_blocks = ceil((double)n / num_threads);

    trans_cuda<<<num_blocks, num_threads>>>(n, d_val_csr, d_idx, d_val_csc);
    cudaDeviceSynchronize();
}

__global__ void WarpLevel_spgemm_32(int n,
                                    int layer,
                                    int *d_bin_rowpointer,
                                    int *d_bin_rowindex,
                                    int *d_rowPtrA,
                                    int *d_colIdxA,
                                    double *d_valueA,
                                    int *d_rowPtrB,
                                    int *d_colIdxB,
                                    double *d_valueB,
                                    int *d_rowPtrC,
                                    int *d_colIdxC,
                                    double *d_valueC)
{
    int warp_global_id = (blockDim.x * blockIdx.x + threadIdx.x) / 32;
    int warp_local_id = threadIdx.x / 32;
    int lane_id = threadIdx.x % 32;

    if (warp_global_id >= d_bin_rowpointer[layer + 1] - d_bin_rowpointer[layer])
        return;

    const int rowidx = d_bin_rowindex[d_bin_rowpointer[layer] + warp_global_id];

    int therowc = d_rowPtrC[rowidx];
    int nextrowc = d_rowPtrC[rowidx + 1];

    int therowa = d_rowPtrA[rowidx];
    int nextrowa = d_rowPtrA[rowidx + 1];

    int SM_LEN = 32;

    __shared__ int s_idxC[32 * WARP_PER_BLOCK_GEMM];
    __shared__ double s_valC[32 * WARP_PER_BLOCK_GEMM];

    int *s_idxC_local = &s_idxC[warp_local_id * SM_LEN];
    double *s_valC_local = &s_valC[warp_local_id * SM_LEN];

    if (lane_id < nextrowc - therowc)
    {
        s_idxC_local[lane_id] = d_colIdxC[therowc + lane_id];
        s_valC_local[lane_id] = 0;
    }

    for (int i = therowa + lane_id; i < nextrowa; i += WARP_SIZE)
    {
        int colA = d_colIdxA[i];
        double valA = d_valueA[i];

        int therowb = d_rowPtrB[colA];
        int nextrowb = d_rowPtrB[colA + 1];

        for (int j = therowb; j < nextrowb; j++)
        {
            int colB = d_colIdxB[j];
            int flag = binarySearch(s_idxC_local, 0, nextrowc - therowc - 1, colB);
            atomicAdd(&s_valC_local[flag], -valA * d_valueB[j]);
        }
    }

    if (lane_id < nextrowc - therowc)
    {
        d_valueC[therowc + lane_id] += s_valC_local[lane_id];
    }
}

__global__ void WarpLevel_spgemm_64(int n,
                                    int layer,
                                    int *d_bin_rowpointer,
                                    int *d_bin_rowindex,
                                    int *d_rowPtrA,
                                    int *d_colIdxA,
                                    double *d_valueA,
                                    int *d_rowPtrB,
                                    int *d_colIdxB,
                                    double *d_valueB,
                                    int *d_rowPtrC,
                                    int *d_colIdxC,
                                    double *d_valueC)
{
    int warp_global_id = (blockDim.x * blockIdx.x + threadIdx.x) / 32;
    int warp_local_id = threadIdx.x / 32;
    int lane_id = threadIdx.x % 32;

    if (warp_global_id >= d_bin_rowpointer[layer + 1] - d_bin_rowpointer[layer])
        return;

    const int rowidx = d_bin_rowindex[d_bin_rowpointer[layer] + warp_global_id];

    int therowc = d_rowPtrC[rowidx];
    int nextrowc = d_rowPtrC[rowidx + 1];

    int therowa = d_rowPtrA[rowidx];
    int nextrowa = d_rowPtrA[rowidx + 1];

    int SM_LEN = 64;

    __shared__ int s_idxC[64 * WARP_PER_BLOCK_GEMM];
    __shared__ double s_valC[64 * WARP_PER_BLOCK_GEMM];

    int *s_idxC_local = &s_idxC[warp_local_id * SM_LEN];
    double *s_valC_local = &s_valC[warp_local_id * SM_LEN];

    for (int i = lane_id; i < nextrowc - therowc; i += WARP_SIZE)
    {
        s_idxC_local[i] = d_colIdxC[therowc + i];
        s_valC_local[i] = 0;
    }
    __syncthreads();

    for (int i = therowa + lane_id; i < nextrowa; i += WARP_SIZE)
    {
        int colA = d_colIdxA[i];
        double valA = d_valueA[i];

        int therowb = d_rowPtrB[colA];
        int nextrowb = d_rowPtrB[colA + 1];

        for (int j = therowb; j < nextrowb; j++)
        {
            int colB = d_colIdxB[j];
            int flag = binarySearch(s_idxC_local, 0, nextrowc - therowc - 1, colB);
            atomicAdd(&s_valC_local[flag], -valA * d_valueB[j]);
        }
    }
    for (int i = lane_id; i < nextrowc - therowc; i += WARP_SIZE)
    {
        d_valueC[therowc + i] += s_valC_local[i];
    }
    __syncthreads();
}

__global__ void WarpLevel_spgemm_128(int n,
                                     int layer,
                                     int *d_bin_rowpointer,
                                     int *d_bin_rowindex,
                                     int *d_rowPtrA,
                                     int *d_colIdxA,
                                     double *d_valueA,
                                     int *d_rowPtrB,
                                     int *d_colIdxB,
                                     double *d_valueB,
                                     int *d_rowPtrC,
                                     int *d_colIdxC,
                                     double *d_valueC)
{
    int warp_global_id = (blockDim.x * blockIdx.x + threadIdx.x) / 32;
    int warp_local_id = threadIdx.x / 32;
    int lane_id = threadIdx.x % 32;

    if (warp_global_id >= d_bin_rowpointer[layer + 1] - d_bin_rowpointer[layer])
        return;

    const int rowidx = d_bin_rowindex[d_bin_rowpointer[layer] + warp_global_id];

    int therowc = d_rowPtrC[rowidx];
    int nextrowc = d_rowPtrC[rowidx + 1];

    int therowa = d_rowPtrA[rowidx];
    int nextrowa = d_rowPtrA[rowidx + 1];

    int SM_LEN = 128;

    __shared__ int s_idxC[128 * WARP_PER_BLOCK_GEMM];
    __shared__ double s_valC[128 * WARP_PER_BLOCK_GEMM];

    int *s_idxC_local = &s_idxC[warp_local_id * SM_LEN];
    double *s_valC_local = &s_valC[warp_local_id * SM_LEN];

    for (int i = lane_id; i < nextrowc - therowc; i += WARP_SIZE)
    {
        s_idxC_local[i] = d_colIdxC[therowc + i];
        s_valC_local[i] = 0;
    }
    __syncthreads();

    for (int i = therowa; i < nextrowa; i++)
    {
        int colA = d_colIdxA[i];
        double valA = d_valueA[i];

        int therowb = d_rowPtrB[colA];
        int nextrowb = d_rowPtrB[colA + 1];

        for (int j = therowb + lane_id; j < nextrowb; j += WARP_SIZE)
        {
            int colB = d_colIdxB[j];
            int flag = binarySearch(s_idxC_local, 0, nextrowc - therowc - 1, colB);
            s_valC_local[flag] -= valA * d_valueB[j];
        }
    }

    for (int i = lane_id; i < nextrowc - therowc; i += WARP_SIZE)
    {
        d_valueC[therowc + i] += s_valC_local[i];
    }
    __syncthreads();
}

__global__ void ThreadLevel_spgemm(int n,
                                   int layer,
                                   int *d_bin_rowpointer,
                                   int *d_bin_rowindex,
                                   int *d_rowPtrA,
                                   int *d_colIdxA,
                                   double *d_valueA,
                                   int *d_rowPtrB,
                                   int *d_colIdxB,
                                   double *d_valueB,
                                   int *d_rowPtrC,
                                   int *d_colIdxC,
                                   double *d_valueC)
{
    const int rowidx = d_bin_rowindex[d_bin_rowpointer[layer] + blockDim.x * blockIdx.x + threadIdx.x];

    if (d_bin_rowpointer[layer] + blockDim.x * blockIdx.x + threadIdx.x >= d_bin_rowpointer[layer + 1])
        return;

    int therowa = d_rowPtrA[rowidx];
    int nextrowa = d_rowPtrA[rowidx + 1];

    int therowc = d_rowPtrC[rowidx];

    for (int i = therowa; i < nextrowa; i++)
    {
        int colA = d_colIdxA[i];
        double valA = d_valueA[i];

        int therowb = d_rowPtrB[colA];
        int nextrowb = d_rowPtrB[colA + 1];

        if (nextrowb - therowb != 0)
            d_valueC[therowc] -= valA * d_valueB[therowb];
    }
    __syncthreads();
}

__global__ void BlockLevel_spgemm_256(int n,
                                      int layer,
                                      int *d_bin_rowpointer,
                                      int *d_bin_rowindex,
                                      int *d_rowPtrA,
                                      int *d_colIdxA,
                                      double *d_valueA,
                                      int *d_rowPtrB,
                                      int *d_colIdxB,
                                      double *d_valueB,
                                      int *d_rowPtrC,
                                      int *d_colIdxC,
                                      double *d_valueC)
{
    int warp_local_id = threadIdx.x / 32;
    int warp_num = blockDim.x / 32;
    int lane_id = threadIdx.x % 32;

    const int rowidx = d_bin_rowindex[d_bin_rowpointer[layer] + blockIdx.x];

    __shared__ int s_idxC[256];
    __shared__ double s_valC[256];

    int therowc = d_rowPtrC[rowidx];
    int nextrowc = d_rowPtrC[rowidx + 1];

    for (int i = threadIdx.x; i < nextrowc - therowc; i += blockDim.x)
    {
        s_idxC[i] = d_colIdxC[therowc + i];
        s_valC[i] = 0;
    }
    __syncthreads();

    int therow = d_rowPtrA[rowidx];
    int nextrow = d_rowPtrA[rowidx + 1];

    for (int i = therow + warp_local_id; i < nextrow; i += warp_num)
    {
        int colA = d_colIdxA[i];
        double valA = d_valueA[i];

        int therowb = d_rowPtrB[colA];
        int nextrowb = d_rowPtrB[colA + 1];

        for (int j = therowb + lane_id; j < nextrowb; j += WARP_SIZE)
        {
            int colB = d_colIdxB[j];
            int flag = binarySearch(s_idxC, 0, nextrowc - therowc - 1, colB);
            atomicAdd(&s_valC[flag], -valA * d_valueB[j]);
        }
    }
    __syncthreads();

    for (int i = threadIdx.x; i < nextrowc - therowc; i += blockDim.x)
    {
        d_valueC[therowc + i] += s_valC[i];
    }
}

__global__ void BlockLevel_spgemm_512(int n,
                                      int layer,
                                      int *d_bin_rowpointer,
                                      int *d_bin_rowindex,
                                      int *d_rowPtrA,
                                      int *d_colIdxA,
                                      double *d_valueA,
                                      int *d_rowPtrB,
                                      int *d_colIdxB,
                                      double *d_valueB,
                                      int *d_rowPtrC,
                                      int *d_colIdxC,
                                      double *d_valueC)
{
    int warp_local_id = threadIdx.x / 32;
    int warp_num = blockDim.x / 32;
    int lane_id = threadIdx.x % 32;

    const int rowidx = d_bin_rowindex[d_bin_rowpointer[layer] + blockIdx.x];

    __shared__ int s_idxC[512];
    __shared__ double s_valC[512];

    int therowc = d_rowPtrC[rowidx];
    int nextrowc = d_rowPtrC[rowidx + 1];

    for (int i = threadIdx.x; i < nextrowc - therowc; i += blockDim.x)
    {
        s_idxC[i] = d_colIdxC[therowc + i];
        s_valC[i] = 0;
    }
    __syncthreads();

    int therow = d_rowPtrA[rowidx];
    int nextrow = d_rowPtrA[rowidx + 1];

    for (int i = therow + warp_local_id; i < nextrow; i += warp_num)
    {
        int colA = d_colIdxA[i];
        double valA = d_valueA[i];

        int therowb = d_rowPtrB[colA];
        int nextrowb = d_rowPtrB[colA + 1];

        for (int j = therowb + lane_id; j < nextrowb; j += WARP_SIZE)
        {
            int colB = d_colIdxB[j];
            int flag = binarySearch(s_idxC, 0, nextrowc - therowc - 1, colB);
            atomicAdd(&s_valC[flag], -valA * d_valueB[j]);
        }
    }
    __syncthreads();

    for (int i = threadIdx.x; i < nextrowc - therowc; i += blockDim.x)
    {
        d_valueC[therowc + i] += s_valC[i];
    }
}

__global__ void BlockLevel_spgemm_1024(int n,
                                       int layer,
                                       int *d_bin_rowpointer,
                                       int *d_bin_rowindex,
                                       int *d_rowPtrA,
                                       int *d_colIdxA,
                                       double *d_valueA,
                                       int *d_rowPtrB,
                                       int *d_colIdxB,
                                       double *d_valueB,
                                       int *d_rowPtrC,
                                       int *d_colIdxC,
                                       double *d_valueC)
{
    int warp_local_id = threadIdx.x / 32;
    int warp_num = blockDim.x / 32;
    int lane_id = threadIdx.x % 32;

    const int rowidx = d_bin_rowindex[d_bin_rowpointer[layer] + blockIdx.x];

    __shared__ int s_idxC[1024];
    __shared__ double s_valC[1024];

    int therowc = d_rowPtrC[rowidx];
    int nextrowc = d_rowPtrC[rowidx + 1];

    for (int i = threadIdx.x; i < nextrowc - therowc; i += blockDim.x)
    {
        s_idxC[i] = d_colIdxC[therowc + i];
        s_valC[i] = 0;
    }
    __syncthreads();

    int therow = d_rowPtrA[rowidx];
    int nextrow = d_rowPtrA[rowidx + 1];

    for (int i = therow + warp_local_id; i < nextrow; i += warp_num)
    {
        int colA = d_colIdxA[i];
        double valA = d_valueA[i];

        int therowb = d_rowPtrB[colA];
        int nextrowb = d_rowPtrB[colA + 1];

        for (int j = therowb + lane_id; j < nextrowb; j += WARP_SIZE)
        {
            int colB = d_colIdxB[j];
            int flag = binarySearch(s_idxC, 0, nextrowc - therowc - 1, colB);
            atomicAdd(&s_valC[flag], -valA * d_valueB[j]);
        }
    }
    __syncthreads();

    for (int i = threadIdx.x; i < nextrowc - therowc; i += blockDim.x)
    {
        d_valueC[therowc + i] += s_valC[i];
    }
}

__global__ void BlockLevel_spgemm_2048(int n,
                                       int layer,
                                       int *d_bin_rowpointer,
                                       int *d_bin_rowindex,
                                       int *d_rowPtrA,
                                       int *d_colIdxA,
                                       double *d_valueA,
                                       int *d_rowPtrB,
                                       int *d_colIdxB,
                                       double *d_valueB,
                                       int *d_rowPtrC,
                                       int *d_colIdxC,
                                       double *d_valueC)
{
    int warp_local_id = threadIdx.x / 32;
    int warp_num = blockDim.x / 32;
    int lane_id = threadIdx.x % 32;

    const int rowidx = d_bin_rowindex[d_bin_rowpointer[layer] + blockIdx.x];

    __shared__ int s_idxC[2048];
    __shared__ double s_valC[2048];

    int therowc = d_rowPtrC[rowidx];
    int nextrowc = d_rowPtrC[rowidx + 1];

    for (int i = threadIdx.x; i < nextrowc - therowc; i += blockDim.x)
    {
        s_idxC[i] = d_colIdxC[therowc + i];
        s_valC[i] = 0;
    }
    __syncthreads();

    int therow = d_rowPtrA[rowidx];
    int nextrow = d_rowPtrA[rowidx + 1];

    for (int i = therow + warp_local_id; i < nextrow; i += warp_num)
    {
        int colA = d_colIdxA[i];
        double valA = d_valueA[i];

        int therowb = d_rowPtrB[colA];
        int nextrowb = d_rowPtrB[colA + 1];

        for (int j = therowb + lane_id; j < nextrowb; j += WARP_SIZE)
        {
            int colB = d_colIdxB[j];
            int flag = binarySearch(s_idxC, 0, nextrowc - therowc - 1, colB);
            atomicAdd(&s_valC[flag], -valA * d_valueB[j]);
        }
    }
    __syncthreads();

    for (int i = threadIdx.x; i < nextrowc - therowc; i += blockDim.x)
    {
        d_valueC[therowc + i] += s_valC[i];
    }
}

__global__ void BlockLevel_spgemm_4096(int n,
                                       int layer,
                                       int *d_bin_rowpointer,
                                       int *d_bin_rowindex,
                                       int *d_rowPtrA,
                                       int *d_colIdxA,
                                       double *d_valueA,
                                       int *d_rowPtrB,
                                       int *d_colIdxB,
                                       double *d_valueB,
                                       int *d_rowPtrC,
                                       int *d_colIdxC,
                                       double *d_valueC)
{
    int warp_local_id = threadIdx.x / 32;
    int warp_num = blockDim.x / 32;
    int lane_id = threadIdx.x % 32;

    const int rowidx = d_bin_rowindex[d_bin_rowpointer[layer] + blockIdx.x];

    __shared__ int s_idxC[4096];
    __shared__ double s_valC[4096];

    int therowc = d_rowPtrC[rowidx];
    int nextrowc = d_rowPtrC[rowidx + 1];

    for (int i = threadIdx.x; i < nextrowc - therowc; i += blockDim.x)
    {
        s_idxC[i] = d_colIdxC[therowc + i];
        s_valC[i] = 0;
    }
    __syncthreads();

    int therow = d_rowPtrA[rowidx];
    int nextrow = d_rowPtrA[rowidx + 1];

    for (int i = therow + warp_local_id; i < nextrow; i += warp_num)
    {
        int colA = d_colIdxA[i];
        double valA = d_valueA[i];

        int therowb = d_rowPtrB[colA];
        int nextrowb = d_rowPtrB[colA + 1];

        for (int j = therowb + lane_id; j < nextrowb; j += WARP_SIZE)
        {
            int colB = d_colIdxB[j];
            int flag = binarySearch(s_idxC, 0, nextrowc - therowc - 1, colB);
            atomicAdd(&s_valC[flag], -valA * d_valueB[j]);
        }
    }
    __syncthreads();

    for (int i = threadIdx.x; i < nextrowc - therowc; i += blockDim.x)
    {
        d_valueC[therowc + i] += s_valC[i];
    }
}

__global__ void BlockLevel_spgemm_4097(int n,
                                       int layer,
                                       int *d_bin_rowpointer,
                                       int *d_bin_rowindex,
                                       int *d_rowPtrA,
                                       int *d_colIdxA,
                                       double *d_valueA,
                                       int *d_rowPtrB,
                                       int *d_colIdxB,
                                       double *d_valueB,
                                       int *d_rowPtrC,
                                       int *d_colIdxC,
                                       double *d_valueC)
{
    int warp_local_id = threadIdx.x / 32;
    int warp_num = blockDim.x / 32;
    int lane_id = threadIdx.x % 32;

    const int rowidx = d_bin_rowindex[d_bin_rowpointer[layer] + blockIdx.x];

    int therow = d_rowPtrA[rowidx];
    int nextrow = d_rowPtrA[rowidx + 1];

    int therowc = d_rowPtrC[rowidx];
    int nextrowc = d_rowPtrC[rowidx + 1];

    for (int i = therow + warp_local_id; i < nextrow; i += warp_num)
    {
        int colA = d_colIdxA[i];
        double valA = d_valueA[i];

        int therowb = d_rowPtrB[colA];
        int nextrowb = d_rowPtrB[colA + 1];

        for (int j = therowb + lane_id; j < nextrowb; j += WARP_SIZE)
        {
            int colB = d_colIdxB[j];

            int flag = binarySearch(d_colIdxC, therowc, nextrowc - 1, colB);

            atomicAdd(&d_valueC[flag], -valA * d_valueB[j]);
        }

        __syncthreads();
    }
}

void pangulu_ssssm_cuda_kernel(int n,
                               int *h_bin_rowpointer,
                               int *d_bin_rowpointer,
                               int *d_bin_rowindex,
                               int *d_rowPtrA,
                               int *d_colIdxA,
                               double *d_valueA,
                               int *d_rowPtrB,
                               int *d_colIdxB,
                               double *d_valueB,
                               int *d_rowPtrC,
                               int *d_colIdxC,
                               double *d_valueC)
{

    int num_blocks;
    int num_threads;
    int layer;

    // 0:0
    layer = 0;

    // 1:1
    layer = 1;
    num_threads = WARP_PER_BLOCK_GEMM * WARP_SIZE;
    num_blocks = ceil((double)(h_bin_rowpointer[2] - h_bin_rowpointer[1]) / num_threads);
    if (num_blocks > 0)
    {
        ThreadLevel_spgemm<<<num_blocks, num_threads>>>(n, layer, d_bin_rowpointer, d_bin_rowindex,
                                                        d_rowPtrA, d_colIdxA, d_valueA,
                                                        d_rowPtrB, d_colIdxB, d_valueB,
                                                        d_rowPtrC, d_colIdxC, d_valueC);
    }

    // 2:2-32
    layer = 2;
    num_threads = WARP_PER_BLOCK_GEMM * WARP_SIZE;
    num_blocks = ceil((double)(h_bin_rowpointer[3] - h_bin_rowpointer[2]) / WARP_PER_BLOCK_GEMM);
    if (num_blocks > 0)
    {
        WarpLevel_spgemm_32<<<num_blocks, num_threads>>>(n, layer, d_bin_rowpointer, d_bin_rowindex,
                                                         d_rowPtrA, d_colIdxA, d_valueA,
                                                         d_rowPtrB, d_colIdxB, d_valueB,
                                                         d_rowPtrC, d_colIdxC, d_valueC);
    }

    // 3:33-64
    layer = 3;
    num_threads = WARP_PER_BLOCK_GEMM * WARP_SIZE;
    num_blocks = ceil((double)(h_bin_rowpointer[4] - h_bin_rowpointer[3]) / WARP_PER_BLOCK_GEMM);
    if (num_blocks > 0)
    {
        WarpLevel_spgemm_64<<<num_blocks, num_threads>>>(n, layer, d_bin_rowpointer, d_bin_rowindex,
                                                         d_rowPtrA, d_colIdxA, d_valueA,
                                                         d_rowPtrB, d_colIdxB, d_valueB,
                                                         d_rowPtrC, d_colIdxC, d_valueC);
    }

    // 4:65-128
    layer = 4;
    num_threads = WARP_PER_BLOCK_GEMM * WARP_SIZE;
    num_blocks = ceil((double)(h_bin_rowpointer[5] - h_bin_rowpointer[4]) / WARP_PER_BLOCK_GEMM);
    if (num_blocks > 0)
    {
        WarpLevel_spgemm_128<<<num_blocks, num_threads>>>(n, layer, d_bin_rowpointer, d_bin_rowindex,
                                                          d_rowPtrA, d_colIdxA, d_valueA,
                                                          d_rowPtrB, d_colIdxB, d_valueB,
                                                          d_rowPtrC, d_colIdxC, d_valueC);
    }

    // 5:129-256
    layer = 5;
    num_threads = 64;
    num_blocks = h_bin_rowpointer[6] - h_bin_rowpointer[5];
    if (num_blocks > 0)
    {
        BlockLevel_spgemm_256<<<num_blocks, num_threads>>>(n, layer, d_bin_rowpointer, d_bin_rowindex,
                                                           d_rowPtrA, d_colIdxA, d_valueA,
                                                           d_rowPtrB, d_colIdxB, d_valueB,
                                                           d_rowPtrC, d_colIdxC, d_valueC);
    }

    // 6:257-512
    layer = 6;
    num_threads = 128;
    num_blocks = h_bin_rowpointer[7] - h_bin_rowpointer[6];
    if (num_blocks > 0)
    {
        BlockLevel_spgemm_512<<<num_blocks, num_threads>>>(n, layer, d_bin_rowpointer, d_bin_rowindex,
                                                           d_rowPtrA, d_colIdxA, d_valueA,
                                                           d_rowPtrB, d_colIdxB, d_valueB,
                                                           d_rowPtrC, d_colIdxC, d_valueC);
    }

    // 7:513-1024
    layer = 7;
    num_threads = 256;
    num_blocks = h_bin_rowpointer[8] - h_bin_rowpointer[7];
    if (num_blocks > 0)
    {
        BlockLevel_spgemm_1024<<<num_blocks, num_threads>>>(n, layer, d_bin_rowpointer, d_bin_rowindex,
                                                            d_rowPtrA, d_colIdxA, d_valueA,
                                                            d_rowPtrB, d_colIdxB, d_valueB,
                                                            d_rowPtrC, d_colIdxC, d_valueC);
    }

    // 8:1025-2048
    layer = 8;
    num_threads = 512;
    num_blocks = h_bin_rowpointer[9] - h_bin_rowpointer[8];
    if (num_blocks > 0)
    {
        BlockLevel_spgemm_2048<<<num_blocks, num_threads>>>(n, layer, d_bin_rowpointer, d_bin_rowindex,
                                                            d_rowPtrA, d_colIdxA, d_valueA,
                                                            d_rowPtrB, d_colIdxB, d_valueB,
                                                            d_rowPtrC, d_colIdxC, d_valueC);
    }

    // 9:2049-4096
    layer = 9;
    num_threads = 1024;
    num_blocks = h_bin_rowpointer[10] - h_bin_rowpointer[9];
    if (num_blocks > 0)
    {
        BlockLevel_spgemm_4096<<<num_blocks, num_threads>>>(n, layer, d_bin_rowpointer, d_bin_rowindex,
                                                            d_rowPtrA, d_colIdxA, d_valueA,
                                                            d_rowPtrB, d_colIdxB, d_valueB,
                                                            d_rowPtrC, d_colIdxC, d_valueC);
    }

    // 10:4097+++
    layer = 10;
    num_threads = 1024;
    num_blocks = h_bin_rowpointer[11] - h_bin_rowpointer[10];
    if (num_blocks > 0)
    {
        BlockLevel_spgemm_4097<<<num_blocks, num_threads>>>(n, layer, d_bin_rowpointer, d_bin_rowindex,
                                                            d_rowPtrA, d_colIdxA, d_valueA,
                                                            d_rowPtrB, d_colIdxB, d_valueB,
                                                            d_rowPtrC, d_colIdxC, d_valueC);
    }
}

//SpTRSV
__global__ void sptrsv_syncfree_cuda_analyser(const int *d_cscRowIdx,
                                              const int m,
                                              const int nnz,
                                              int *d_graphInDegree)
{
    const int global_id = blockIdx.x * blockDim.x + threadIdx.x; //get_global_id(0);
    if (global_id < nnz)
    {
        atomicAdd(&d_graphInDegree[d_cscRowIdx[global_id]], 1);
    }
}

__global__ void sptrsv_syncfree_cuda_executor_update_L(const int *d_cscColPtr,
                                                       const int *d_cscRowIdx,
                                                       const double *d_cscVal,
                                                       int *d_graphInDegree,
                                                       double *d_left_sum,
                                                       const int m,
                                                       const double *d_b,
                                                       double *d_x,
                                                       int *d_id_extractor)
{
    const int global_id = blockIdx.x * blockDim.x + threadIdx.x;

    // Initialize
    const int local_warp_id = threadIdx.x / WARP_SIZE;
    const int lane_id = (WARP_SIZE - 1) & threadIdx.x;
    int global_x_id = 0;
    if (!lane_id)
        global_x_id = atomicAdd(d_id_extractor, 1);
    global_x_id = __shfl(global_x_id, 0);

    if (global_x_id >= m)
        return;

    // substitution is forward
    global_x_id = global_x_id;

    // Prefetch
    const int pos = d_cscColPtr[global_x_id];
    const double coef = (double)1 / d_cscVal[pos];

    // Consumer
    do
    {
        __threadfence_block();
    } while (d_graphInDegree[global_x_id] != 1);

    double xi = d_left_sum[global_x_id];
    xi = (d_b[global_x_id] - xi) * coef;

    // Producer
    const int start_ptr = d_cscColPtr[global_x_id] + 1;
    const int stop_ptr = d_cscColPtr[global_x_id + 1];
    for (int jj = start_ptr + lane_id; jj < stop_ptr; jj += WARP_SIZE)
    {
        const int j = jj;
        const int rowIdx = d_cscRowIdx[j];

        atomicAdd(&d_left_sum[rowIdx], xi * d_cscVal[j]);
        __threadfence();
        atomicSub(&d_graphInDegree[rowIdx], 1);
    }

    //finish
    if (!lane_id)
        d_x[global_x_id] = xi;
}
__global__ void sptrsv_syncfree_cuda_executor_update_U(const int *d_cscColPtr,
                                                       const int *d_cscRowIdx,
                                                       const double *d_cscVal,
                                                       int *d_graphInDegree,
                                                       double *d_left_sum,
                                                       const int m,
                                                       const double *d_b,
                                                       double *d_x,
                                                       int *d_id_extractor)
{
    const int global_id = blockIdx.x * blockDim.x + threadIdx.x;

    // Initialize
    const int local_warp_id = threadIdx.x / WARP_SIZE;
    const int lane_id = (WARP_SIZE - 1) & threadIdx.x;
    int global_x_id = 0;
    if (!lane_id)
        global_x_id = atomicAdd(d_id_extractor, 1);
    global_x_id = __shfl(global_x_id, 0);

    if (global_x_id >= m)
        return;

    // substitution is backward
    global_x_id = m - 1 - global_x_id;

    // Prefetch
    const int pos = d_cscColPtr[global_x_id + 1] - 1;
    const double coef = (double)1 / d_cscVal[pos];

    // Consumer
    do
    {
        __threadfence_block();
    } while (d_graphInDegree[global_x_id] != 1);

    double xi = d_left_sum[global_x_id];
    xi = (d_b[global_x_id] - xi) * coef;

    // Producer
    const int start_ptr = d_cscColPtr[global_x_id];
    const int stop_ptr = d_cscColPtr[global_x_id + 1] - 1;
    for (int jj = start_ptr + lane_id; jj < stop_ptr; jj += WARP_SIZE)
    {
        const int j = stop_ptr - 1 - (jj - start_ptr);
        const int rowIdx = d_cscRowIdx[j];

        atomicAdd(&d_left_sum[rowIdx], xi * d_cscVal[j]);
        __threadfence();
        atomicSub(&d_graphInDegree[rowIdx], 1);
    }

    //finish
    if (!lane_id)
        d_x[global_x_id] = xi;
}

__forceinline__ __device__ double sum_32_shfl(double sum)
{
#pragma unroll
    for (int mask = WARP_SIZE / 2; mask > 0; mask >>= 1)
        sum += __shfl_xor_sync(0xffffffff, sum, mask);

    return sum;
}
__global__ void spmv_warpvec_csr_cuda_executor(const int *d_csrRowPtr,
                                               const int *d_csrColIdx,
                                               const double *d_csrVal,
                                               const int m,
                                               const double *d_x,
                                               double *d_y)
{
    const int global_id = blockIdx.x * blockDim.x + threadIdx.x;

    // Initialize
    const int rowid = global_id / WARP_SIZE;
    if (rowid >= m)
        return;

    const int lane_id = (WARP_SIZE - 1) & threadIdx.x;
    const int start = d_csrRowPtr[rowid];
    const int stop = d_csrRowPtr[rowid + 1];
    if (start == stop)
    {
        if (!lane_id)
            d_y[rowid] = 0;
        return;
    }

    double sum = 0;

    for (int j = start + lane_id; j < stop; j += WARP_SIZE)
    {
        sum += d_x[d_csrColIdx[j]] * d_csrVal[j];
    }
    sum = sum_32_shfl(sum);

    //finish
    if (!lane_id)
        d_y[rowid] = sum;
}

void pangulu_sptrsv_dispose_L_kernel(int m,
                                     int n,
                                     int nnzTR,
                                     int *CUDA_cscColPtrTR,
                                     int *CUDA_cscRowIdxTR,
                                     double *CUDA_cscValTR,
                                     double *CUDA_x,
                                     double *CUDA_b,
                                     int *d_graphInDegree,
                                     int *d_graphInDegree_backup,
                                     double *d_left_sum,
                                     int *d_id_extractor)
{

    int num_threads = 128;
    //int num_threads = n;
    int num_blocks = ceil((double)nnzTR / (double)num_threads);

    cudaMemset(d_graphInDegree, 0, m * sizeof(int));
    sptrsv_syncfree_cuda_analyser<<<num_blocks, num_threads>>>(CUDA_cscRowIdxTR, m, nnzTR, d_graphInDegree);

    cudaDeviceSynchronize();
    cudaMemcpy(d_graphInDegree_backup, d_graphInDegree, m * sizeof(int), cudaMemcpyDeviceToDevice);

    // get a unmodified in-degree array, only for benchmarking use
    cudaMemcpy(d_graphInDegree, d_graphInDegree_backup, m * sizeof(int), cudaMemcpyDeviceToDevice);
    //cudaMemset(d_graphInDegree, 0, sizeof(int) * m);

    // clear left_sum array, only for benchmarking use
    cudaMemset(d_left_sum, 0, sizeof(double) * m);

    cudaMemset(d_id_extractor, 0, sizeof(int));

    num_threads = WARP_PER_BLOCK * WARP_SIZE;
    //num_threads = 64;
    //num_threads = 1 * WARP_SIZE;
    num_blocks = ceil((double)m / (double)(num_threads / WARP_SIZE));
    //sptrsv_syncfree_cuda_executor<<< num_blocks, num_threads >>>

    sptrsv_syncfree_cuda_executor_update_L<<<num_blocks, num_threads>>>(CUDA_cscColPtrTR, CUDA_cscRowIdxTR, CUDA_cscValTR,
                                                                        d_graphInDegree, d_left_sum,
                                                                        m, CUDA_b, CUDA_x, d_id_extractor);

    cudaDeviceSynchronize();
}
void pangulu_sptrsv_dispose_U_kernel(int m,
                                     int n,
                                     int nnzTR,
                                     int *CUDA_cscColPtrTR,
                                     int *CUDA_cscRowIdxTR,
                                     double *CUDA_cscValTR,
                                     double *CUDA_x,
                                     double *CUDA_b,
                                     int *d_graphInDegree,
                                     int *d_graphInDegree_backup,
                                     double *d_left_sum,
                                     int *d_id_extractor)
{

    int num_threads = 128;
    //int num_threads = n;
    int num_blocks = ceil((double)nnzTR / (double)num_threads);

    cudaMemset(d_graphInDegree, 0, m * sizeof(int));
    sptrsv_syncfree_cuda_analyser<<<num_blocks, num_threads>>>(CUDA_cscRowIdxTR, m, nnzTR, d_graphInDegree);

    cudaDeviceSynchronize();
    cudaMemcpy(d_graphInDegree_backup, d_graphInDegree, m * sizeof(int), cudaMemcpyDeviceToDevice);
    // get a unmodified in-degree array, only for benchmarking use

    cudaMemcpy(d_graphInDegree, d_graphInDegree_backup, m * sizeof(int), cudaMemcpyDeviceToDevice);
    //cudaMemset(d_graphInDegree, 0, sizeof(int) * m);

    // clear left_sum array, only for benchmarking use
    cudaMemset(d_left_sum, 0, sizeof(double) * m);

    cudaMemset(d_id_extractor, 0, sizeof(int));

    num_threads = WARP_PER_BLOCK * WARP_SIZE;
    //num_threads = 1 * WARP_SIZE;

    num_blocks = ceil((double)m / (double)(num_threads / WARP_SIZE));

    //sptrsv_syncfree_cuda_executor<<< num_blocks, num_threads >>>
    sptrsv_syncfree_cuda_executor_update_U<<<num_blocks, num_threads>>>(CUDA_cscColPtrTR, CUDA_cscRowIdxTR, CUDA_cscValTR,
                                                                        d_graphInDegree, d_left_sum,
                                                                        m, CUDA_b, CUDA_x, d_id_extractor);

    cudaDeviceSynchronize();
}

void pangulu_sptrsv_spmv_kernel(int n,
                                int *CUDA_rowpointer,
                                int *CUDA_columnindex,
                                double *CUDA_VALUE,
                                double *CUDA_x,
                                double *CUDA_y)
{

    int num_threads = WARP_PER_BLOCK * WARP_SIZE;

    int num_blocks = ceil((double)n / (double)(num_threads / WARP_SIZE));

    spmv_warpvec_csr_cuda_executor<<<num_blocks, num_threads>>>(CUDA_rowpointer, CUDA_columnindex, CUDA_VALUE, n, CUDA_x, CUDA_y);
    cudaDeviceSynchronize();
}

__global__ void GESSM_Kernel_v2(int n,
                                int *L_columnpointer,
                                int *L_rowindex,
                                double *L_VALUE,
                                int *X_columnpointer,
                                int *X_rowindex,
                                double *X_VALUE,
                                int *A_columnpointer,
                                int *A_rowindex,
                                double *A_VALUE)
{
    int warp_id = (blockDim.x * blockIdx.x + threadIdx.x) / 32;
    int lane_id = threadIdx.x % 32;
    int warp_local_id = threadIdx.x / 32;
    int flagx, flagb;

    if (warp_id >= n || X_columnpointer[warp_id] == X_columnpointer[warp_id + 1])
        return;

    int colA1 = A_columnpointer[warp_id];
    int colA2 = A_columnpointer[warp_id + 1];
    int colX1 = X_columnpointer[warp_id];
    int colX2 = X_columnpointer[warp_id + 1];

    if (colA2 - colA1 >= SM_LEN_DGESSM)
    {
        for (int i = colX1, t = 0; i < colX2; i++, t++)
        {
            int rowX = X_rowindex[i];
            X_VALUE[i] = A_VALUE[i];
            double valx = X_VALUE[i];

            int colL1 = L_columnpointer[rowX];
            int colL2 = L_columnpointer[rowX + 1];
            for (int j = colL1 + 1 + lane_id, p = lane_id; j < colL2; j += WARP_SIZE, p += WARP_SIZE)
            {
                int f = binarySearch(A_rowindex, colA1 + 1 + t + p, colA2 - colL2 + j, L_rowindex[j]);
                A_VALUE[f] -= valx * L_VALUE[j];
            }
            __syncthreads();
        }
    }
    else
    {
        __shared__ int s_rowidxA[SM_LEN_DGESSM * WARP_PER_BLOCK_DGESSM];
        __shared__ double s_valueA[SM_LEN_DGESSM * WARP_PER_BLOCK_DGESSM];

        int *idx_local = &s_rowidxA[warp_local_id * SM_LEN_DGESSM];
        double *val_local = &s_valueA[warp_local_id * SM_LEN_DGESSM];

        for (int i = lane_id; i < colA2 - colA1; i += WARP_SIZE)
        {
            idx_local[i] = A_rowindex[colA1 + i];
            val_local[i] = A_VALUE[colA1 + i];
        }

        for (int i = colX1, t = 0; i < colX2; i++, t++)
        {
            int rowX = X_rowindex[i];
            X_VALUE[i] = val_local[t];
            double valx = X_VALUE[i];

            int colL1 = L_columnpointer[rowX];
            int colL2 = L_columnpointer[rowX + 1];
            for (int j = colL1 + 1 + lane_id, p = lane_id; j < colL2; j += WARP_SIZE, p += WARP_SIZE)
            {
                //update A's Value;
                int f = binarySearch(idx_local, 1 + t + p, colA2 - colA1 - colL2 + j, L_rowindex[j]);
                if (f != -1)
                    val_local[f] -= valx * L_VALUE[j];
            }
            __syncthreads();
        }

        for (int i = lane_id; i < colA2 - colA1; i += WARP_SIZE)
        {
            A_VALUE[colA1 + i] = val_local[i];
        }
    }
}

__global__ void TSTRF_Kernel_v2(int n,
                                int *U_rowpointer,
                                int *U_columnindex,
                                double *U_VALUE,
                                int *X_rowpointer,
                                int *X_columnindex,
                                double *X_VALUE,
                                int *A_rowpointer,
                                int *A_columnindex,
                                double *A_VALUE)
{
    int warp_id = (blockDim.x * blockIdx.x + threadIdx.x) / 32;
    int lane_id = threadIdx.x % 32;
    int warp_local_id = threadIdx.x / 32;

    if (warp_id >= n || X_rowpointer[warp_id] == X_rowpointer[warp_id + 1])
        return;

    int colA1 = A_rowpointer[warp_id];
    int colA2 = A_rowpointer[warp_id + 1];
    int colX1 = X_rowpointer[warp_id];
    int colX2 = X_rowpointer[warp_id + 1];

    if (colA2 - colA1 >= SM_LEN_DTSTRF)
    {
        for (int i = colX1, t = 0; i < colX2; i++, t++)
        {
            int rowX = X_columnindex[i];
            int colU1 = U_rowpointer[rowX];
            int colU2 = U_rowpointer[rowX + 1];

            X_VALUE[i] = A_VALUE[i] / U_VALUE[colU1];
            double valx = X_VALUE[i];
            for (int j = colU1 + 1 + lane_id, p = lane_id; j < colU2; j += WARP_SIZE, p += WARP_SIZE)
            {
                //update A's Value;
                int f = binarySearch(A_columnindex, colA1 + t + 1 + p, colA2 - colU2 + j, U_columnindex[j]);
                A_VALUE[f] -= valx * U_VALUE[j];
            }
            __syncthreads();
        }
    }
    else
    {
        __shared__ int s_rowidxA[SM_LEN_DTSTRF * WARP_PER_BLOCK_DTSTRF];
        __shared__ double s_valueA[SM_LEN_DTSTRF * WARP_PER_BLOCK_DTSTRF];

        int *idx_local = &s_rowidxA[warp_local_id * SM_LEN_DTSTRF];
        double *val_local = &s_valueA[warp_local_id * SM_LEN_DTSTRF];

        for (int i = lane_id; i < colA2 - colA1; i += WARP_SIZE)
        {
            idx_local[i] = A_columnindex[colA1 + i];
            val_local[i] = A_VALUE[colA1 + i];
        }

        for (int i = colX1, t = 0; i < colX2; i++, t++)
        {
            int rowX = X_columnindex[i];
            int colU1 = U_rowpointer[rowX];
            int colU2 = U_rowpointer[rowX + 1];

            X_VALUE[i] = val_local[t] / U_VALUE[colU1];
            double valx = X_VALUE[i];
            for (int j = colU1 + 1 + lane_id, p = lane_id; j < colU2; j += WARP_SIZE, p += WARP_SIZE)
            {
                //update A's Value;
                int f = binarySearch(idx_local, 1 + t + p, colA2 - colA1 - colU2 + j, U_columnindex[j]);
                val_local[f] -= valx * U_VALUE[j];
            }
            __syncthreads();
        }

        for (int i = lane_id; i < colA2 - colA1; i += WARP_SIZE)
        {
            A_VALUE[colA1 + i] = val_local[i];
        }
    }
}

__global__ void GESSM_Kernel_v3(int n,
                                int *L_columnpointer,
                                int *L_rowindex,
                                double *L_VALUE,
                                int *X_columnpointer,
                                int *X_rowindex,
                                double *X_VALUE,
                                int *A_columnpointer,
                                int *A_rowindex,
                                double *A_VALUE)
{

    int colidx = blockIdx.x;
    int colX1 = X_columnpointer[colidx];
    int colX2 = X_columnpointer[colidx + 1];

    if (colidx >= n || colX2 == colX1)
        return;

    int colA1 = A_columnpointer[colidx];
    int colA2 = A_columnpointer[colidx + 1];

    if (colA2 - colA1 >= SM_LEN_DGESSM)
    {
        for (int i = colX1, t = 0; i < colX2; i++, t++)
        {
            int rowX = X_rowindex[i];
            X_VALUE[i] = A_VALUE[i];
            double valx = X_VALUE[i];

            int colL1 = L_columnpointer[rowX];
            int colL2 = L_columnpointer[rowX + 1];

            for (int j = colL1 + 1 + threadIdx.x, p = threadIdx.x; j < colL2; j += blockDim.x, p += blockDim.x)
            {
                //update A's Value;
                int f = binarySearch(A_rowindex, colA1 + 1 + t + p, colA2 - colL2 + j, L_rowindex[j]);
                A_VALUE[f] -= valx * L_VALUE[j];
            }
            __syncthreads();
        }
    }
    else
    {
        __shared__ int s_idxA[SM_LEN_DGESSM];
        __shared__ double s_valA[SM_LEN_DGESSM];

        for (int i = threadIdx.x; i < colA2 - colA1; i += blockDim.x)
        {
            s_idxA[i] = A_rowindex[colA1 + i];
            s_valA[i] = A_VALUE[colA1 + i];
        }
        __syncthreads();

        for (int i = colX1, t = 0; i < colX2; i++, t++)
        {
            int rowX = X_rowindex[i];
            X_VALUE[i] = s_valA[t];
            double valx = X_VALUE[i];

            int colL1 = L_columnpointer[rowX];
            int colL2 = L_columnpointer[rowX + 1];

            for (int j = colL1 + 1 + threadIdx.x, p = threadIdx.x; j < colL2; j += blockDim.x, p += blockDim.x)
            {
                //update A's Value;
                int f = binarySearch(s_idxA, 1 + t + p, colA2 - colA1 - colL2 + j, L_rowindex[j]);
                s_valA[f] -= valx * L_VALUE[j];
            }
            __syncthreads();
        }

        for (int i = threadIdx.x; i < colA2 - colA1; i += blockDim.x)
        {
            A_VALUE[colA1 + i] = s_valA[i];
        }
        //__syncthreads();
    }
}

__global__ void TSTRF_Kernel_v3(int n,
                                int *U_rowpointer,
                                int *U_columnindex,
                                double *U_VALUE,
                                int *X_rowpointer,
                                int *X_columnindex,
                                double *X_VALUE,
                                int *A_rowpointer,
                                int *A_columnindex,
                                double *A_VALUE)
{
    int colidx = blockIdx.x;
    int colX1 = X_rowpointer[colidx];
    int colX2 = X_rowpointer[colidx + 1];

    if (colidx >= n || colX1 == colX2)
        return;

    int colA1 = A_rowpointer[colidx];
    int colA2 = A_rowpointer[colidx + 1];

    if (colA2 - colA1 >= SM_LEN_DTSTRF)
    {
        for (int i = colX1, t = 0; i < colX2; i++, t++)
        {
            int rowX = X_columnindex[i];
            int colU1 = U_rowpointer[rowX];
            int colU2 = U_rowpointer[rowX + 1];

            X_VALUE[i] = A_VALUE[i] / U_VALUE[colU1];
            double valx = X_VALUE[i];
            for (int j = colU1 + 1 + threadIdx.x, p = threadIdx.x; j < colU2; j += blockDim.x, p += blockDim.x)
            {
                //update A's Value;
                int f = binarySearch(A_columnindex, colA1 + t + 1 + p, colA2 - colU2 + j, U_columnindex[j]);
                A_VALUE[f] -= valx * U_VALUE[j];
            }
            __syncthreads();
        }
    }
    else
    {
        __shared__ int s_rowidxA[SM_LEN_DTSTRF];
        __shared__ double s_valueA[SM_LEN_DTSTRF];

        for (int i = threadIdx.x; i < colA2 - colA1; i += blockDim.x)
        {
            s_rowidxA[i] = A_columnindex[colA1 + i];
            s_valueA[i] = A_VALUE[colA1 + i];
        }
        __syncthreads();

        for (int i = colX1, t = 0; i < colX2; i++, t++)
        {
            int rowX = X_columnindex[i];
            int colU1 = U_rowpointer[rowX];
            int colU2 = U_rowpointer[rowX + 1];

            X_VALUE[i] = s_valueA[t] / U_VALUE[colU1];
            double valx = X_VALUE[i];
            for (int j = colU1 + 1 + threadIdx.x, p = threadIdx.x; j < colU2; j += blockDim.x, p += blockDim.x)
            {
                //update A's Value;
                int f = binarySearch(s_rowidxA, 1 + t + p, colA2 - colA1 - colU2 + j, U_columnindex[j]);
                s_valueA[f] -= valx * U_VALUE[j];
            }
            __syncthreads();
        }

        for (int i = threadIdx.x; i < colA2 - colA1; i += blockDim.x)
        {
            A_VALUE[colA1 + i] = s_valueA[i];
        }
    }
}

void pangulu_gessm_cuda_kernel(int n,
                               int nnzL,
                               int *L_columnpointer,
                               int *L_rowindex,
                               double *L_VALUE,
                               int *X_columnpointer,
                               int *X_rowindex,
                               double *X_VALUE,
                               int *A_columnpointer,
                               int *A_rowindex,
                               double *A_VALUE)
{
    int nnzL_avrg = nnzL / n;

    if (nnzL_avrg < 32)
    {
        int num_threads = WARP_PER_BLOCK_DGESSM * WARP_SIZE;
        int num_blocks = ceil((double)n / WARP_PER_BLOCK_DGESSM);
        GESSM_Kernel_v2<<<num_blocks, num_threads>>>(n, L_columnpointer, L_rowindex, L_VALUE,
                                                     X_columnpointer, X_rowindex, X_VALUE,
                                                     A_columnpointer, A_rowindex, A_VALUE);
    }
    else
    {
        int num_threads = 8 * WARP_SIZE;
        int num_blocks = n;
        GESSM_Kernel_v3<<<num_blocks, num_threads>>>(n, L_columnpointer, L_rowindex, L_VALUE,
                                                     X_columnpointer, X_rowindex, X_VALUE,
                                                     A_columnpointer, A_rowindex, A_VALUE);
    }
}

void pangulu_tstrf_cuda_kernel(int n,
                               int nnzU,
                               int *U_rowpointer,
                               int *U_columnindex,
                               double *U_VALUE,
                               int *X_rowpointer,
                               int *X_columnindex,
                               double *X_VALUE,
                               int *A_rowpointer,
                               int *A_columnindex,
                               double *A_VALUE)
{
    int nnzU_avrg = nnzU / n;
    if (nnzU_avrg < 32)
    {
        int num_threads = WARP_PER_BLOCK_DTSTRF * WARP_SIZE;
        int num_blocks = ceil((double)n / WARP_PER_BLOCK_DTSTRF);
        TSTRF_Kernel_v2<<<num_blocks, num_threads>>>(n, U_rowpointer, U_columnindex, U_VALUE,
                                                     X_rowpointer, X_columnindex, X_VALUE,
                                                     A_rowpointer, A_columnindex, A_VALUE);
    }
    else
    {
        int num_threads = 8 * WARP_SIZE;
        int num_blocks = n;
        TSTRF_Kernel_v3<<<num_blocks, num_threads>>>(n, U_rowpointer, U_columnindex, U_VALUE,
                                                     X_rowpointer, X_columnindex, X_VALUE,
                                                     A_rowpointer, A_columnindex, A_VALUE);
    }
}
