#ifndef PANGULU_CUDA_H
#define PANGULU_CUDA_H
#ifdef __cplusplus
extern "C"
{
#endif
#define WARP_SIZE 32
#define WARP_PER_BLOCK_GEMM 2
#define WARP_NUM_WARPLU 2
#define SM_LEN_WARPLEV 96
#define WARP_PER_BLOCK 8
#define LONGROW_THRESHOLD 2048
#define SM_LEN_DGESSM 1024
#define SM_LEN_DTSTRF 1024
#define WARP_PER_BLOCK_DGESSM 2
#define WARP_PER_BLOCK_DTSTRF 2
#include "pangulu_common.h"
#include <cuda_runtime.h>
    /*
The function in this file is primarily called kernel,these functions would be called in pangulu_cuda_interface.h
*/
    void pangulu_cuda_malloc_double(double **cuda_address, size_t size);
    double *pangulu_cuda_free_double(double *cuda_address);
    void pangulu_cuda_memcpy_host_to_device_double(double *cpu_address, double *cuda_address, size_t size);
    void pangulu_cuda_memcpy_device_to_host_double(double *cuda_address, double *cpu_address, size_t size);
    void pangulu_cuda_devicesynchronize();
    void pangulu_cuda_malloc_int(int **cuda_address, size_t size);
    int *pangulu_cuda_free_int(int *cuda_address);
    void pangulu_cuda_memcpy_host_to_device_int(int *cuda_address, int *cpu_address, size_t size);
    void pangulu_cuda_memcpy_device_to_host_int(int *cpu_address, int *cuda_address, size_t size);
    void pangulu_cuda_getDevicenum(int *gpu_num);
    int pangulu_cuda_setDevice(int gpu_num, int rank);

    __global__ void WarpLevel_sflu(int n,
                                   int *d_nnzU,
                                   int *d_cscColPtrA,
                                   int *d_cscRowIdxA,
                                   double *d_cscValueA,
                                   int *d_cscColPtrL,
                                   int *d_cscRowIdxL,
                                   double *d_cscValueL,
                                   int *d_cscColPtrU,
                                   int *d_cscRowIdxU,
                                   double *d_cscValueU);

    __global__ void BlockLevel_sflu_L1(int n,
                                       int *d_nnzU,
                                       int *d_cscColPtrA,
                                       int *d_cscRowIdxA,
                                       double *d_cscValueA,
                                       int *d_cscColPtrL,
                                       int *d_cscRowIdxL,
                                       double *d_cscValueL,
                                       int *d_cscColPtrU,
                                       int *d_cscRowIdxU,
                                       double *d_cscValueU);

    __global__ void BlockLevel_sflu_L2(int n,
                                       int *d_nnzU,
                                       int *d_cscColPtrA,
                                       int *d_cscRowIdxA,
                                       double *d_cscValueA,
                                       int *d_cscColPtrL,
                                       int *d_cscRowIdxL,
                                       double *d_cscValueL,
                                       int *d_cscColPtrU,
                                       int *d_cscRowIdxU,
                                       double *d_cscValueU);

    __global__ void BlockLevel_sflu_L3(int n,
                                       int *d_nnzU,
                                       int *d_cscColPtrA,
                                       int *d_cscRowIdxA,
                                       double *d_cscValueA,
                                       int *d_cscColPtrL,
                                       int *d_cscRowIdxL,
                                       double *d_cscValueL,
                                       int *d_cscColPtrU,
                                       int *d_cscRowIdxU,
                                       double *d_cscValueU);

    __global__ void BlockLevel_sflu_L4(int n,
                                       int *d_nnzU,
                                       int *d_cscColPtrA,
                                       int *d_cscRowIdxA,
                                       double *d_cscValueA,
                                       int *d_cscColPtrL,
                                       int *d_cscRowIdxL,
                                       double *d_cscValueL,
                                       int *d_cscColPtrU,
                                       int *d_cscRowIdxU,
                                       double *d_cscValueU);

    int binarySearch(int *ridx, int left, int right, int target);

    void pangulu_getrf_cuda_kernel(int n,
                                   int nnzA,
                                   int *d_nnzU,
                                   int *A_CUDA_rowpointer,
                                   int *A_CUDA_columnindex,
                                   double *A_CUDA_value,
                                   int *L_CUDA_rowpointer,
                                   int *L_CUDA_columnindex,
                                   double *L_CUDA_value,
                                   int *U_CUDA_rowpointer,
                                   int *U_CUDA_columnindex,
                                   double *U_CUDA_value);

    void pangulu_cuda_transport_kernel(int n, double *d_val_csr, int *d_idx, double *d_val_csc);

    __global__ void trans_cuda(int n, double *d_val_csr, int *d_idx, double *d_val_csc);

    __global__ void ThreadLevel_spgemm(int n,
                                       int layer,
                                       int *d_bin_rowpointer,
                                       int *d_bin_rowindex,
                                       int *d_rowPtrA,
                                       int *d_colIdxA,
                                       double *d_valueA,
                                       int *d_rowPtrB,
                                       int *d_colIdxB,
                                       double *d_valueB,
                                       int *d_rowPtrC,
                                       int *d_colIdxC,
                                       double *d_valueC);

    __global__ void WarpLevel_spgemm_32(int n,
                                        int layer,
                                        int *d_bin_rowpointer,
                                        int *d_bin_rowindex,
                                        int *d_rowPtrA,
                                        int *d_colIdxA,
                                        double *d_valueA,
                                        int *d_rowPtrB,
                                        int *d_colIdxB,
                                        double *d_valueB,
                                        int *d_rowPtrC,
                                        int *d_colIdxC,
                                        double *d_valueC);

    __global__ void WarpLevel_spgemm_64(int n,
                                        int layer,
                                        int *d_bin_rowpointer,
                                        int *d_bin_rowindex,
                                        int *d_rowPtrA,
                                        int *d_colIdxA,
                                        double *d_valueA,
                                        int *d_rowPtrB,
                                        int *d_colIdxB,
                                        double *d_valueB,
                                        int *d_rowPtrC,
                                        int *d_colIdxC,
                                        double *d_valueC);

    __global__ void WarpLevel_spgemm_128(int n,
                                         int layer,
                                         int *d_bin_rowpointer,
                                         int *d_bin_rowindex,
                                         int *d_rowPtrA,
                                         int *d_colIdxA,
                                         double *d_valueA,
                                         int *d_rowPtrB,
                                         int *d_colIdxB,
                                         double *d_valueB,
                                         int *d_rowPtrC,
                                         int *d_colIdxC,
                                         double *d_valueC);

    __global__ void BlockLevel_spgemm_256(int n,
                                          int layer,
                                          int *d_bin_rowpointer,
                                          int *d_bin_rowindex,
                                          int *d_rowPtrA,
                                          int *d_colIdxA,
                                          double *d_valueA,
                                          int *d_rowPtrB,
                                          int *d_colIdxB,
                                          double *d_valueB,
                                          int *d_rowPtrC,
                                          int *d_colIdxC,
                                          double *d_valueC);

    __global__ void BlockLevel_spgemm_512(int n,
                                          int layer,
                                          int *d_bin_rowpointer,
                                          int *d_bin_rowindex,
                                          int *d_rowPtrA,
                                          int *d_colIdxA,
                                          double *d_valueA,
                                          int *d_rowPtrB,
                                          int *d_colIdxB,
                                          double *d_valueB,
                                          int *d_rowPtrC,
                                          int *d_colIdxC,
                                          double *d_valueC);

    __global__ void BlockLevel_spgemm_1024(int n,
                                           int layer,
                                           int *d_bin_rowpointer,
                                           int *d_bin_rowindex,
                                           int *d_rowPtrA,
                                           int *d_colIdxA,
                                           double *d_valueA,
                                           int *d_rowPtrB,
                                           int *d_colIdxB,
                                           double *d_valueB,
                                           int *d_rowPtrC,
                                           int *d_colIdxC,
                                           double *d_valueC);

    __global__ void BlockLevel_spgemm_2048(int n,
                                           int layer,
                                           int *d_bin_rowpointer,
                                           int *d_bin_rowindex,
                                           int *d_rowPtrA,
                                           int *d_colIdxA,
                                           double *d_valueA,
                                           int *d_rowPtrB,
                                           int *d_colIdxB,
                                           double *d_valueB,
                                           int *d_rowPtrC,
                                           int *d_colIdxC,
                                           double *d_valueC);

    __global__ void BlockLevel_spgemm_4096(int n,
                                           int layer,
                                           int *d_bin_rowpointer,
                                           int *d_bin_rowindex,
                                           int *d_rowPtrA,
                                           int *d_colIdxA,
                                           double *d_valueA,
                                           int *d_rowPtrB,
                                           int *d_colIdxB,
                                           double *d_valueB,
                                           int *d_rowPtrC,
                                           int *d_colIdxC,
                                           double *d_valueC);

    __global__ void BlockLevel_spgemm_4097(int n,
                                           int layer,
                                           int *d_bin_rowpointer,
                                           int *d_bin_rowindex,
                                           int *d_rowPtrA,
                                           int *d_colIdxA,
                                           double *d_valueA,
                                           int *d_rowPtrB,
                                           int *d_colIdxB,
                                           double *d_valueB,
                                           int *d_rowPtrC,
                                           int *d_colIdxC,
                                           double *d_valueC);

    void pangulu_ssssm_cuda_kernel(int n,
                                   int *h_bin_rowpointer,
                                   int *d_bin_rowpointer,
                                   int *d_bin_rowindex,
                                   int *d_rowPtrA,
                                   int *d_colIdxA,
                                   double *d_valueA,
                                   int *d_rowPtrB,
                                   int *d_colIdxB,
                                   double *d_valueB,
                                   int *d_rowPtrC,
                                   int *d_colIdxC,
                                   double *d_valueC);

    __global__ void sptrsv_syncfree_cuda_analyser(const int *d_cscRowIdx,
                                                  const int m,
                                                  const int nnz,
                                                  int *d_graphInDegree);

    __global__ void sptrsv_syncfree_cuda_executor_update_U(const int *d_cscColPtr,
                                                           const int *d_cscRowIdx,
                                                           const double *d_cscVal,
                                                           int *d_graphInDegree,
                                                           double *d_left_sum,
                                                           const int m,
                                                           const double *d_b,
                                                           double *d_x,
                                                           int *d_id_extractor);

    __global__ void sptrsv_syncfree_cuda_executor_update_L(const int *d_cscColPtr,
                                                           const int *d_cscRowIdx,
                                                           const double *d_cscVal,
                                                           int *d_graphInDegree,
                                                           double *d_left_sum,
                                                           const int m,
                                                           const double *d_b,
                                                           double *d_x,
                                                           int *d_id_extractor);

    __forceinline__ __device__ double sum_32_shfl(double sum);

    __global__ void spmv_warpvec_csr_cuda_executor(const int *d_csrRowPtr,
                                                   const int *d_csrColIdx,
                                                   const double *d_csrVal,
                                                   const int m,
                                                   const double *d_x,
                                                   double *d_y);

    void pangulu_sptrsv_dispose_L_kernel(int m,
                                         int n,
                                         int nnzTR,
                                         int *CUDA_cscColPtrTR,
                                         int *CUDA_cscRowIdxTR,
                                         double *CUDA_cscValTR,
                                         double *CUDA_x,
                                         double *CUDA_b,
                                         int *d_graphInDegree,
                                         int *d_graphInDegree_backup,
                                         double *d_left_sum,
                                         int *d_id_extractor);

    void pangulu_sptrsv_dispose_U_kernel(int m,
                                         int n,
                                         int nnzTR,
                                         int *CUDA_cscColPtrTR,
                                         int *CUDA_cscRowIdxTR,
                                         double *CUDA_cscValTR,
                                         double *CUDA_x,
                                         double *CUDA_b,
                                         int *d_graphInDegree,
                                         int *d_graphInDegree_backup,
                                         double *d_left_sum,
                                         int *d_id_extractor);

    void pangulu_sptrsv_spmv_kernel(int n,
                                    int *CUDA_rowpointer,
                                    int *CUDA_columnindex,
                                    double *CUDA_VALUE,
                                    double *CUDA_x,
                                    double *CUDA_y);

    __global__ void GESSM_Kernel_v2(int n,
                                    int *L_columnpointer,
                                    int *L_rowindex,
                                    double *L_VALUE,
                                    int *X_columnpointer,
                                    int *X_rowindex,
                                    double *X_VALUE,
                                    int *A_columnpointer,
                                    int *A_rowindex,
                                    double *A_VALUE);

    __global__ void TSTRF_Kernel_v2(int n,
                                    int *U_rowpointer,
                                    int *U_columnindex,
                                    double *U_VALUE,
                                    int *X_rowpointer,
                                    int *X_columnindex,
                                    double *X_VALUE,
                                    int *A_rowpointer,
                                    int *A_columnindex,
                                    double *A_VALUE);

    __global__ void GESSM_Kernel_v3(int n,
                                    int *L_columnpointer,
                                    int *L_rowindex,
                                    double *L_VALUE,
                                    int *X_columnpointer,
                                    int *X_rowindex,
                                    double *X_VALUE,
                                    int *A_columnpointer,
                                    int *A_rowindex,
                                    double *A_VALUE);

    __global__ void TSTRF_Kernel_v3(int n,
                                    int *U_rowpointer,
                                    int *U_columnindex,
                                    double *U_VALUE,
                                    int *X_rowpointer,
                                    int *X_columnindex,
                                    double *X_VALUE,
                                    int *A_rowpointer,
                                    int *A_columnindex,
                                    double *A_VALUE);

    void pangulu_gessm_cuda_kernel(int n,
                                   int nnzL, int *L_columnpointer,
                                   int *L_rowindex,
                                   double *L_VALUE,
                                   int *X_columnpointer,
                                   int *X_rowindex,
                                   double *X_VALUE,
                                   int *A_columnpointer,
                                   int *A_rowindex,
                                   double *A_VALUE);

    void pangulu_tstrf_cuda_kernel(int n,
                                   int nnzU,
                                   int *U_rowpointer,
                                   int *U_columnindex,
                                   double *U_VALUE,
                                   int *X_rowpointer,
                                   int *X_columnindex,
                                   double *X_VALUE,
                                   int *A_rowpointer,
                                   int *A_columnindex,
                                   double *A_VALUE);

#ifdef __cplusplus
}
#endif

#endif