#ifndef PANGULU_NUMERICAL_H
#define PANGULU_NUMERICAL_H
#ifdef __cplusplus
extern "C"
{
#endif
#include "pangulu_common.h"
#include "pangulu_kernel_interface.h"
    /*An impoortant function is pangulu_numerical(int rank,
                           int offset_row,
                           int offset_col,
                           int block_length,
                           int P,
                           int Q,
                           int NB,
                           int rank_row_length,
                           int rank_col_length,
                           int *mapper_Big_pangulu_Smatrix,
                           pangulu_Smatrix **Big_pangulu_Smatrix_value,
                           int *L_pangulu_Smatrix_columnpointer,
                           int *L_pangulu_Smatrix_rowindex,
                           pangulu_Smatrix **L_pangulu_Smatrix_value,
                           int *U_pangulu_Smatrix_rowpointer,
                           int *U_pangulu_Smatrix_columnindex,
                           pangulu_Smatrix **U_pangulu_Smatrix_value){ }
It's parameters are explained as follows: 
rank is this process number;
offset_row is The row offset relative to the initial position in the current iteration.
offset_col is The col offset relative to the initial position in the current iteration.
P is the number of process grid rows;
Q is the number of process grid columns;
NB is the size of the sub-matrix block;
rank_row_length is the row length of the current process matrix.
rank_col_length is the col length of the current process matrix.
block_length is the number of rows of the matrix after division;
mapper_Big_pangulu_Smatrix is the mapping relationship of matrix A.
Big_pangulu_Smatrix_value is element value of A matrix;
L_pangulu_Smatrix_columnpointer is the column offset of the L matrix;
L_pangulu_Smatrix_rowindex is the subscript of the non-zero elements row of L matrix;
L_pangulu_Smatrix_value is element value of L matrix;
U_pangulu_Smatrix_rowpointer is the row offset of the U matrix;
U_pangulu_Smatrix_columnindex is the subscript of the non-zero elements column of U matrix;
U_pangulu_Smatrix_value is element value of U matrix.
*/
    void pangulu_solve_pangulu_Smatrix_A_pangulu_Smatrix_L_and_pangulu_Smatrix_U(int rank,
                                                                                 int level,
                                                                                 int block_length,
                                                                                 int P,
                                                                                 int Q,
                                                                                 int NB,
                                                                                 pangulu_Smatrix *A,
                                                                                 pangulu_Smatrix *L,
                                                                                 pangulu_Smatrix *U)
    {
        pangulu_getrf_interface(A, L, U);
        int send_id = rank;
        int signal = level * block_length + level;
        for (int i = 0; i < ((Q - 1) < (block_length - level - 1) ? (Q - 1) : (block_length - level - 1)); i++)
        {
            if ((send_id + 1) % Q == 0)
            {
                send_id -= Q;
            }
            send_id += 1;
            pangulu_send_pangulu_Smatrix_value(L, send_id, signal, NB);
        }
        send_id = rank;
        for (int i = 0; i < ((P - 1) < (block_length - level - 1) ? (P - 1) : (block_length - level - 1)); i++)
        {
            send_id += Q;
            if (send_id >= P * Q)
            {
                send_id = send_id % Q;
            }
            pangulu_send_pangulu_Smatrix_value(U, send_id, signal, NB);
        }
    }

    void pangulu_solve_panglu_Smatrix_L_col_block_panglu_Smatrix(int rank,
                                                                 int level,
                                                                 int block_length,
                                                                 int P,
                                                                 int Q,
                                                                 int NB,
                                                                 int signal,
                                                                 pangulu_Smatrix *A,
                                                                 pangulu_Smatrix *U,
                                                                 pangulu_Smatrix *X)
    {
        pangulu_tstrf_interface(A, X, U);
        //do the kernel X*U=A A[mapper_A[index_value]]
        int send_id = rank;
        for (int i = 0; i < ((Q - 1) < ((block_length - level - 1)) ? (Q - 1) : (block_length - level - 1)); i++)
        {
            if ((send_id + 1) % Q == 0)
            {
                send_id -= Q;
            }
            send_id += 1;
            pangulu_send_pangulu_Smatrix_value(X, send_id, signal, NB);
        }
        pangulu_cuda_memcpy_value_host_to_device(X);
    }

    void pangulu_solve_panglu_Smatrix_U_row_block_panglu_Smatrix(int rank,
                                                                 int level,
                                                                 int block_length,
                                                                 int P,
                                                                 int Q,
                                                                 int NB,
                                                                 int signal,
                                                                 pangulu_Smatrix *A,
                                                                 pangulu_Smatrix *L,
                                                                 pangulu_Smatrix *X)
    {
        pangulu_gessm_interface(A, L, X);
        //do the kernel L*X=A A[mapper_A[index_value]]
        int send_id = rank;
        for (int i = 0; i < ((P - 1) < (block_length - level - 1) ? (P - 1) : (block_length - level - 1)); i++)
        {
            send_id += Q;
            if (send_id >= P * Q)
            {
                send_id = send_id % Q;
            }
            pangulu_send_pangulu_Smatrix_value(X, send_id, signal, NB);
        }
        pangulu_cuda_memcpy_value_host_to_device(X);
    }

    void pangulu_level_solve(int rank,
                             int offset_row,
                             int offset_col,
                             int level,
                             int block_length,
                             int P,
                             int Q,
                             int NB,
                             int rank_size,
                             int *mapper_Big_pangulu_Smatrix,
                             pangulu_Smatrix **Big_pangulu_Smatrix,
                             int L_col_nnz,
                             int U_row_nnz,
                             int rank_row_length,
                             int rank_col_length,
                             int now_row_length,
                             int now_col_length,
                             int *L_pangulu_Smatrix_rowindex,
                             pangulu_Smatrix **L_pangulu_Smatrix_value,
                             int *U_pangulu_Smatrix_columnindex,
                             pangulu_Smatrix **U_pangulu_Smatrix_value)
    {

        if (offset_row == 0 && offset_col == 0)
        {
            int level_index = pangulu_mapper_to_Big_pangulu_Smatrix(level,
                                                                    level,
                                                                    mapper_Big_pangulu_Smatrix,
                                                                    rank_col_length, P, Q);
            pangulu_solve_pangulu_Smatrix_A_pangulu_Smatrix_L_and_pangulu_Smatrix_U(rank,
                                                                                    level,
                                                                                    block_length,
                                                                                    P,
                                                                                    Q,
                                                                                    NB,
                                                                                    Big_pangulu_Smatrix[level_index],
                                                                                    L_pangulu_Smatrix_value[0],
                                                                                    U_pangulu_Smatrix_value[0]);
            int block_row = 1, block_col = 1;
            for (; block_col < L_col_nnz && block_row < U_row_nnz; block_row++, block_col++)
            {
                int row = L_pangulu_Smatrix_rowindex[block_col];
                int mapper_col_index = pangulu_mapper_to_Big_pangulu_Smatrix(row,
                                                                             level,
                                                                             mapper_Big_pangulu_Smatrix,
                                                                             rank_col_length, P, Q);
                pangulu_solve_panglu_Smatrix_L_col_block_panglu_Smatrix(rank,
                                                                        level,
                                                                        block_length,
                                                                        P,
                                                                        Q,
                                                                        NB,
                                                                        level + block_length * row,
                                                                        Big_pangulu_Smatrix[mapper_col_index],
                                                                        U_pangulu_Smatrix_value[0],
                                                                        L_pangulu_Smatrix_value[block_col]);

                int col = U_pangulu_Smatrix_columnindex[block_row];
                int mapper_row_index = pangulu_mapper_to_Big_pangulu_Smatrix(level,
                                                                             col,
                                                                             mapper_Big_pangulu_Smatrix,
                                                                             rank_col_length, P, Q);
                pangulu_solve_panglu_Smatrix_U_row_block_panglu_Smatrix(rank,
                                                                        level,
                                                                        block_length,
                                                                        P,
                                                                        Q,
                                                                        NB,
                                                                        level * block_length + col,
                                                                        Big_pangulu_Smatrix[mapper_row_index],
                                                                        L_pangulu_Smatrix_value[0],
                                                                        U_pangulu_Smatrix_value[block_row]);
            }
            if (L_col_nnz < U_row_nnz)
            {
                for (; block_row < U_row_nnz; block_row++)
                {
                    int col = U_pangulu_Smatrix_columnindex[block_row];
                    int mapper_row_index = pangulu_mapper_to_Big_pangulu_Smatrix(level,
                                                                                 col,
                                                                                 mapper_Big_pangulu_Smatrix,
                                                                                 rank_col_length, P, Q);
                    pangulu_solve_panglu_Smatrix_U_row_block_panglu_Smatrix(rank,
                                                                            level,
                                                                            block_length,
                                                                            P,
                                                                            Q,
                                                                            NB,
                                                                            level * block_length + col,
                                                                            Big_pangulu_Smatrix[mapper_row_index],
                                                                            L_pangulu_Smatrix_value[0],
                                                                            U_pangulu_Smatrix_value[block_row]);
                }
            }
            else
            {
                for (; block_col < L_col_nnz; block_col++)
                {
                    int row = L_pangulu_Smatrix_rowindex[block_col];
                    int mapper_col_index = pangulu_mapper_to_Big_pangulu_Smatrix(row,
                                                                                 level,
                                                                                 mapper_Big_pangulu_Smatrix,
                                                                                 rank_col_length, P, Q);
                    pangulu_solve_panglu_Smatrix_L_col_block_panglu_Smatrix(rank,
                                                                            level,
                                                                            block_length,
                                                                            P,
                                                                            Q,
                                                                            NB,
                                                                            level + block_length * row,
                                                                            Big_pangulu_Smatrix[mapper_col_index],
                                                                            U_pangulu_Smatrix_value[0],
                                                                            L_pangulu_Smatrix_value[block_col]);
                }
            }
            for (int i = 1; i < U_row_nnz; i++)
            {
                for (int j = 1; j < L_col_nnz; j++)
                {
                    int row = L_pangulu_Smatrix_rowindex[j];
                    int col = U_pangulu_Smatrix_columnindex[i];
                    //do kernel 4
                    int mapper_index = pangulu_mapper_to_Big_pangulu_Smatrix(row,
                                                                             col,
                                                                             mapper_Big_pangulu_Smatrix,
                                                                             rank_col_length, P, Q);
                    if (mapper_index == -1)
                    {
                        continue;
                    }
                    pangulu_ssssm_interface(Big_pangulu_Smatrix[mapper_index],
                                            L_pangulu_Smatrix_value[j],
                                            U_pangulu_Smatrix_value[i]);
                }
            }
            return;
        }

        if (offset_col != 0 && offset_row == 0 && ((level + offset_col) < block_length))
        {
            int receive_id = rank - offset_col;
            if (receive_id < (rank - rank % Q))
            {
                receive_id += Q;
            }
            pangulu_recv_pangulu_Smatrix_value(L_pangulu_Smatrix_value[0], receive_id,
                                               level * block_length + level, NB);
            for (int block_row = 0; block_row < U_row_nnz; block_row++)
            {
                int col = U_pangulu_Smatrix_columnindex[block_row];
                int mapper_row_index = pangulu_mapper_to_Big_pangulu_Smatrix(level,
                                                                             col,
                                                                             mapper_Big_pangulu_Smatrix,
                                                                             rank_col_length, P, Q);
                pangulu_solve_panglu_Smatrix_U_row_block_panglu_Smatrix(rank,
                                                                        level,
                                                                        block_length,
                                                                        P,
                                                                        Q,
                                                                        NB,
                                                                        level * block_length + col,
                                                                        Big_pangulu_Smatrix[mapper_row_index],
                                                                        L_pangulu_Smatrix_value[0],
                                                                        U_pangulu_Smatrix_value[block_row]);
            }
            int small_row = rank - rank % Q;
            int receive_rank = rank - offset_col;
            if (receive_rank < (small_row))
            {
                receive_rank += Q;
            }
            for (int j = 1; j < L_col_nnz; j++)
            {
                int signal = level + block_length * L_pangulu_Smatrix_rowindex[j];
                pangulu_recv_pangulu_Smatrix_value(L_pangulu_Smatrix_value[j], receive_rank,
                                                   signal, NB);
                for (int i = 0; i < U_row_nnz; i++)
                {
                    int row = L_pangulu_Smatrix_rowindex[j];
                    int col = U_pangulu_Smatrix_columnindex[i];
                    //do kernel 4
                    int mapper_index = pangulu_mapper_to_Big_pangulu_Smatrix(row,
                                                                             col,
                                                                             mapper_Big_pangulu_Smatrix,
                                                                             rank_col_length, P, Q);

                    if (mapper_index == -1)
                    {
                        continue;
                    }
                    pangulu_ssssm_interface(Big_pangulu_Smatrix[mapper_index],
                                            L_pangulu_Smatrix_value[j],
                                            U_pangulu_Smatrix_value[i]);
                }
            }
            return;
        }

        if (offset_row != 0 && offset_col == 0 && ((level + offset_row) < block_length))
        {

            int receive_id = rank - offset_row * Q;
            if (receive_id < (rank % Q))
            {
                receive_id += P * Q;
            }
            pangulu_recv_pangulu_Smatrix_value(U_pangulu_Smatrix_value[0], receive_id,
                                               level * block_length + level, NB);

            for (int block_col = 0; block_col < L_col_nnz; block_col++)
            {
                int row = L_pangulu_Smatrix_rowindex[block_col];
                int mapper_col_index = pangulu_mapper_to_Big_pangulu_Smatrix(row,
                                                                             level,
                                                                             mapper_Big_pangulu_Smatrix,
                                                                             rank_col_length, P, Q);
                pangulu_solve_panglu_Smatrix_L_col_block_panglu_Smatrix(rank,
                                                                        level,
                                                                        block_length,
                                                                        P,
                                                                        Q,
                                                                        NB,
                                                                        level + block_length * row,
                                                                        Big_pangulu_Smatrix[mapper_col_index],
                                                                        U_pangulu_Smatrix_value[0],
                                                                        L_pangulu_Smatrix_value[block_col]);
            }

            int small_col = rank % Q;
            int receive_rank = rank - offset_row * Q;

            if (receive_rank < (small_col))
            {
                receive_rank += Q * P;
            }

            for (int i = 1; i < U_row_nnz; i++)
            {

                int signal = U_pangulu_Smatrix_columnindex[i] + block_length * level;
                pangulu_recv_pangulu_Smatrix_value(U_pangulu_Smatrix_value[i], receive_rank,
                                                   signal, NB);
                for (int j = 0; j < L_col_nnz; j++)
                {
                    int row = L_pangulu_Smatrix_rowindex[j];
                    int col = U_pangulu_Smatrix_columnindex[i];
                    //do kernel 4
                    int mapper_index = pangulu_mapper_to_Big_pangulu_Smatrix(row,
                                                                             col,
                                                                             mapper_Big_pangulu_Smatrix,
                                                                             rank_col_length, P, Q);
                    if (mapper_index == -1)
                    {
                        continue;
                    }
                    pangulu_ssssm_interface(Big_pangulu_Smatrix[mapper_index],
                                            L_pangulu_Smatrix_value[j],
                                            U_pangulu_Smatrix_value[i]);
                }
            }

            return;
        }

        if (offset_col != 0 && offset_row != 0)
        {

            int small_col = rank % Q;
            int receive_rank_row = rank - offset_row * Q;

            if (receive_rank_row < (small_col))
            {
                receive_rank_row += Q * P;
            }

            int small_row = rank - rank % Q;
            int receive_rank_col = rank - offset_col;
            if (receive_rank_col < (small_row))
            {
                receive_rank_col += Q;
            }

            int block_level = 0;
            int min_length_LU = ((U_row_nnz > L_col_nnz) ? L_col_nnz : U_row_nnz);
            int mapper_index;

            for (; block_level < min_length_LU; block_level++)
            {

                int signal = L_pangulu_Smatrix_rowindex[block_level] * block_length + level;
                pangulu_recv_pangulu_Smatrix_value(L_pangulu_Smatrix_value[block_level], receive_rank_col,
                                                   signal, NB);
                for (int i = 0; i < block_level; i++)
                {
                    int row = L_pangulu_Smatrix_rowindex[block_level];
                    int col = U_pangulu_Smatrix_columnindex[i];
                    //kernel 4
                    mapper_index = pangulu_mapper_to_Big_pangulu_Smatrix(row,
                                                                         col,
                                                                         mapper_Big_pangulu_Smatrix,
                                                                         rank_col_length, P, Q);
                    if (mapper_index == -1)
                    {
                        continue;
                    }
                    pangulu_ssssm_interface(Big_pangulu_Smatrix[mapper_index],
                                            L_pangulu_Smatrix_value[block_level],
                                            U_pangulu_Smatrix_value[i]);
                }

                signal = U_pangulu_Smatrix_columnindex[block_level] + block_length * level;
                pangulu_recv_pangulu_Smatrix_value(U_pangulu_Smatrix_value[block_level], receive_rank_row,
                                                   signal, NB);
                for (int j = 0; j < block_level; j++)
                {

                    int row = L_pangulu_Smatrix_rowindex[j];
                    int col = U_pangulu_Smatrix_columnindex[block_level];
                    mapper_index = pangulu_mapper_to_Big_pangulu_Smatrix(row,
                                                                         col,
                                                                         mapper_Big_pangulu_Smatrix,
                                                                         rank_col_length, P, Q);
                    if (mapper_index == -1)
                    {
                        continue;
                    }
                    pangulu_ssssm_interface(Big_pangulu_Smatrix[mapper_index],
                                            L_pangulu_Smatrix_value[j],
                                            U_pangulu_Smatrix_value[block_level]);
                    //kernel 4
                }

                int row = L_pangulu_Smatrix_rowindex[block_level];
                int col = U_pangulu_Smatrix_columnindex[block_level];
                //kernel 4
                mapper_index = pangulu_mapper_to_Big_pangulu_Smatrix(row,
                                                                     col,
                                                                     mapper_Big_pangulu_Smatrix,
                                                                     rank_col_length, P, Q);
                if (mapper_index == -1)
                {
                    continue;
                }
                pangulu_ssssm_interface(Big_pangulu_Smatrix[mapper_index],
                                        L_pangulu_Smatrix_value[block_level],
                                        U_pangulu_Smatrix_value[block_level]);
            }

            if (U_row_nnz < L_col_nnz)
            {
                for (; block_level < L_col_nnz; block_level++)
                {
                    int signal = L_pangulu_Smatrix_rowindex[block_level] * block_length + level;
                    pangulu_recv_pangulu_Smatrix_value(L_pangulu_Smatrix_value[block_level], receive_rank_col,
                                                       signal, NB);
                    for (int i = 0; i < U_row_nnz; i++)
                    {

                        int row = L_pangulu_Smatrix_rowindex[block_level];
                        int col = U_pangulu_Smatrix_columnindex[i];
                        //kernel 4
                        int mapper_index = pangulu_mapper_to_Big_pangulu_Smatrix(row,
                                                                                 col,
                                                                                 mapper_Big_pangulu_Smatrix,
                                                                                 rank_col_length, P, Q);
                        if (mapper_index == -1)
                        {
                            continue;
                        }
                        pangulu_ssssm_interface(Big_pangulu_Smatrix[mapper_index],
                                                L_pangulu_Smatrix_value[block_level],
                                                U_pangulu_Smatrix_value[i]);
                    }
                }
            }
            else
            {
                for (; block_level < U_row_nnz; block_level++)
                {
                    int signal = U_pangulu_Smatrix_columnindex[block_level] + block_length * level;
                    //printf("rank is %d receive id is %d signal row is %d\n",rank,receive_rank_row,signal);
                    pangulu_recv_pangulu_Smatrix_value(U_pangulu_Smatrix_value[block_level], receive_rank_row,
                                                       signal, NB);
                    for (int j = 0; j < L_col_nnz; j++)
                    {

                        int row = L_pangulu_Smatrix_rowindex[j];
                        int col = U_pangulu_Smatrix_columnindex[block_level];
                        int mapper_index = pangulu_mapper_to_Big_pangulu_Smatrix(row,
                                                                                 col,
                                                                                 mapper_Big_pangulu_Smatrix,
                                                                                 rank_col_length, P, Q);
                        if (mapper_index == -1)
                        {
                            continue;
                        }
                        pangulu_ssssm_interface(Big_pangulu_Smatrix[mapper_index],
                                                L_pangulu_Smatrix_value[j],
                                                U_pangulu_Smatrix_value[block_level]);
                        //kernel 4
                    }
                }
            }

            return;
        }
    }

    void pangulu_numerical(int rank,
                           int offset_row,
                           int offset_col,
                           int block_length,
                           int P,
                           int Q,
                           int NB,
                           int rank_row_length,
                           int rank_col_length,
                           int *mapper_Big_pangulu_Smatrix,
                           pangulu_Smatrix **Big_pangulu_Smatrix_value,
                           int *L_pangulu_Smatrix_columnpointer,
                           int *L_pangulu_Smatrix_rowindex,
                           pangulu_Smatrix **L_pangulu_Smatrix_value,
                           int *U_pangulu_Smatrix_rowpointer,
                           int *U_pangulu_Smatrix_columnindex,
                           pangulu_Smatrix **U_pangulu_Smatrix_value)
    {

        int although_row = 0;
        int although_col = 0;

        for (int level = 0; level < block_length; level++)
        {
            //LEVEL=level;
            if (rank == 0)
            {
                //printf("level  %d\n",level);
            }
            int new_Smatrix_row_length = (block_length / P + (((block_length % P) > (rank / Q)) ? 1 : 0));
            int new_Smatrix_col_length = (block_length / Q + (((block_length % Q) > (rank % Q)) ? 1 : 0));
            //printf("new row length is %d new col length is %d rank is %d\n",new_Smatrix_row_length,new_Smatrix_col_length,rank);

            if (new_Smatrix_row_length == although_row)
            {
                break;
            }
            if (new_Smatrix_col_length == although_col)
            {
                break;
            }

            int L_col_nzz = L_pangulu_Smatrix_columnpointer[level + 1] - L_pangulu_Smatrix_columnpointer[level];
            int U_row_nzz = U_pangulu_Smatrix_rowpointer[level + 1] - U_pangulu_Smatrix_rowpointer[level];
            pangulu_level_solve(rank,
                                offset_row,
                                offset_col,
                                level,
                                block_length,
                                P,
                                Q,
                                NB,
                                rank,
                                mapper_Big_pangulu_Smatrix,
                                Big_pangulu_Smatrix_value,
                                L_col_nzz, U_row_nzz,
                                rank_row_length,
                                rank_col_length,
                                although_row,
                                although_col,
                                L_pangulu_Smatrix_rowindex + L_pangulu_Smatrix_columnpointer[level],
                                L_pangulu_Smatrix_value + L_pangulu_Smatrix_columnpointer[level],
                                U_pangulu_Smatrix_columnindex + U_pangulu_Smatrix_rowpointer[level],
                                U_pangulu_Smatrix_value + U_pangulu_Smatrix_rowpointer[level]);

            if (offset_row == 0)
            {
                although_row++;
            }
            if (offset_col == 0)
            {
                although_col++;
            }
            offset_row = pangulu_exchange_PQ(offset_row, P);
            offset_col = pangulu_exchange_PQ(offset_col, Q);
        }
    }

#ifdef __cplusplus
}
#endif

#endif