#ifndef PANGULU_SPTRSV_H
#define PANGULU_SPTRSV_H
#ifdef __cplusplus
extern "C"
{
#endif
#include "pangulu_common.h"
#include "pangulu_utils.h"
#include "pangulu_sptrsv_cuda.h"
    /*An impoortant function is void pangulu_sptrsv(int N,
                        int rank,
                        int P,
                        int Q,
                        int NB,
                        int block_length,
                        pangulu_Smatrix *symbolic_Smatrix,
                        int *L_pangulu_Smatrix_columnpointer,
                        int *L_pangulu_Smatrix_rowindex,
                        pangulu_Smatrix **L_pangulu_Smatrix_value,
                        int *U_pangulu_Smatrix_rowpointer,
                        int *U_pangulu_Smatrix_columnindex,
                        pangulu_Smatrix **U_pangulu_Smatrix_value){ }
It's parameters are explained as follows: 
N is the size of the original matrix;
rank is this process number;
P is the number of process grid rows;
Q is the number of process grid columns;
NB is the size of the sub-matrix block;
block_length is the number of rows of the matrix after division;
symbolic_Smatrix is  the matrix that is broken down by symbols;
L_pangulu_Smatrix_columnpointer is the column offset of the L matrix;
L_pangulu_Smatrix_rowindex is the subscript of the non-zero elements row of L matrix;
L_pangulu_Smatrix_value is element value of L matrix;
U_pangulu_Smatrix_rowpointer is the row offset of the U matrix;
U_pangulu_Smatrix_columnindex is the subscript of the non-zero elements column of U matrix;
U_pangulu_Smatrix_value is element value of U matrix.
*/

    void pangulu_sptrsv_spmv(int m,
                             int n,
                             pangulu_Smatrix *A,
                             calculate_type *x,
                             calculate_type *y,
                             calculate_type *spmv_CUDA_x,
                             calculate_type *spmv_CUDA_y)
    {

        pangulu_sptrsv_spmv_inteface(A, m, x, y, spmv_CUDA_x, spmv_CUDA_y);
    }

    void pangulu_sptrsv_dispose_L(int row,
                                  int colnum,
                                  pangulu_Smatrix *A,
                                  calculate_type *x,
                                  calculate_type *b,
                                  calculate_type *CUDA_x,
                                  calculate_type *CUDA_b,
                                  int *d_graphInDegree,
                                  int *d_graphInDegree_backup,
                                  calculate_type *d_left_sum,
                                  int *d_id_extractor)
    {
        pangulu_sptrsv_dispose_L_interface(A,
                                           row,
                                           colnum,
                                           x,
                                           b,
                                           CUDA_x,
                                           CUDA_b,
                                           d_graphInDegree,
                                           d_graphInDegree_backup,
                                           d_left_sum,
                                           d_id_extractor);
    }

    void pangulu_sptrsv_dispose_U(int row,
                                  int colnum,
                                  pangulu_Smatrix *A,
                                  calculate_type *x,
                                  calculate_type *b,
                                  calculate_type *CUDA_x,
                                  calculate_type *CUDA_b,
                                  int *d_graphInDegree,
                                  int *d_graphInDegree_backup,
                                  calculate_type *d_left_sum,
                                  int *d_id_extractor)
    {
        pangulu_sptrsv_dispose_U_interface(A,
                                           row,
                                           colnum,
                                           x,
                                           b,
                                           CUDA_x,
                                           CUDA_b,
                                           d_graphInDegree,
                                           d_graphInDegree_backup,
                                           d_left_sum,
                                           d_id_extractor);
    }

    void pangulu_block_pangulu_Smatrix_CSC_to_block_pangulu_Smatrix_CSR(int rank,
                                                                        int block_length,
                                                                        int P,
                                                                        int Q,
                                                                        int *L_Smatrix_rowpointer,
                                                                        int *L_Smatrix_columnindex,
                                                                        pangulu_Smatrix **L_Smatrix_value,
                                                                        int **L_csr_Smatrix_rowpointer,
                                                                        int **L_csr_Smatrix_columnindex,
                                                                        pangulu_Smatrix ***L_csr_Smatrix_value)
    {
        int *L_save_rowpointer = (int *)malloc(sizeof(int) * (block_length + 1));
        int *L_save_columnindex;
        pangulu_Smatrix **L_save_value;
        for (int i = 0; i < (block_length + 1); i++)
        {
            L_save_rowpointer[i] = 0;
        }
        int Lnzz = 0;
        for (int i = rank % Q; i < block_length; i += Q)
        {
            for (int j = L_Smatrix_rowpointer[i]; j < L_Smatrix_rowpointer[i + 1]; j++)
            {
                int row = L_Smatrix_columnindex[j];
                L_save_rowpointer[row + 1]++;
                Lnzz++;
            }
        }
        for (int i = 0; i < block_length; i++)
        {
            L_save_rowpointer[i + 1] += L_save_rowpointer[i];
        }
        L_save_columnindex = (int *)malloc(sizeof(int) * Lnzz);
        L_save_value = (pangulu_Smatrix **)malloc(sizeof(pangulu_Smatrix *) * Lnzz);
        for (int i = rank % Q; i < block_length; i += Q)
        {
            for (int j = L_Smatrix_rowpointer[i]; j < L_Smatrix_rowpointer[i + 1]; j++)
            {
                int row = L_Smatrix_columnindex[j];
                int index = L_save_rowpointer[row];
                L_save_columnindex[index] = i;
                L_save_value[index] = L_Smatrix_value[j];
                L_save_rowpointer[row]++;
            }
        }
        for (int i = block_length; i > 0; i--)
        {
            L_save_rowpointer[i] = L_save_rowpointer[i - 1];
        }
        L_save_rowpointer[0] = 0;
        *L_csr_Smatrix_rowpointer = L_save_rowpointer;
        *L_csr_Smatrix_columnindex = L_save_columnindex;
        *L_csr_Smatrix_value = L_save_value;
    }

    void pangulu_sptrsv_recv_vector(calculate_type *VALUE, int receive_id, int signal, int NB)
    {
        MPI_Status status;
        MPI_Recv(VALUE, NB, MPI_DOUBLE, receive_id, signal, MPI_COMM_WORLD, &status);
    }

    void pangulu_sptrsv_send_vector(calculate_type *value, int send_id, int signal, int NB)
    {
        MPI_Send(value, NB, MPI_DOUBLE, send_id, signal, MPI_COMM_WORLD);
    }

    double vec2norm(double *x, int n)
    {
        double sum = 0.0;
        for (int i = 0; i < n; i++)
            sum += x[i] * x[i];
        return sqrt(sum);
    }

    calculate_type pangulu_sptrsv_check_calculate_x(calculate_type *x_ref,
                                                    calculate_type **x,
                                                    int block_length,
                                                    int NB,
                                                    int N)
    {
        calculate_type *check_x = (calculate_type *)malloc(sizeof(calculate_type) * N);
        calculate_type e = 0.0;
        int index;
        for (int level = 0; level < block_length; level++)
        {
            for (int i = 0; i < NB; i++)
            {
                index = level * NB + i;
                if (index >= N)
                {
                    e = vec2norm(check_x, N) / vec2norm(x_ref, N);
                    free(check_x);
                    return e;
                }
                else
                {
                    check_x[index] = x_ref[index] - x[level][i];
                }
            }
        }
        e = vec2norm(x_ref, N) / vec2norm(check_x, N);
        free(check_x);
        return e;
    }

    void pangulu_sptrsv_add_b(calculate_type *b, calculate_type *b_now, int NB)
    {
        for (int i = 0; i < NB; i++)
        {
            b[i] += b_now[i];
        }
    }

    void pangulu_sptrsv_changeb(calculate_type *b, calculate_type *b_now, int NB)
    {
        for (int i = 0; i < NB; i++)
        {
            b[i] -= b_now[i];
        }
    }

    void pangulu_sptrsv_dispose_block_pangulu_Smatrix_L(int rank,
                                                        int block_row,
                                                        int block_length,
                                                        int P,
                                                        int Q,
                                                        int NB,
                                                        int N,
                                                        pangulu_Smatrix *L_Smatrix_value,
                                                        calculate_type **b,
                                                        calculate_type **x,
                                                        calculate_type **b_now,
                                                        calculate_type *x_ref,
                                                        calculate_type *b_recieve,
                                                        calculate_type **y1,
                                                        calculate_type *CUDA_x,
                                                        calculate_type *CUDA_b,
                                                        int *d_graphInDegree,
                                                        int *d_graphInDegree_backup,
                                                        calculate_type *d_left_sum,
                                                        int *d_id_extractor)
    {
        int level = block_row;
        int recieve_id;
        for (int recieve_level = ((level - Q + 1) < 0) ? 0 : level - Q + 1; recieve_level < level; recieve_level++)
        {
            recieve_id = recieve_level % Q + level % P * ((level % P) ? Q : 0);
            pangulu_sptrsv_recv_vector(b_recieve, recieve_id, level * block_length + recieve_level, NB);
            pangulu_sptrsv_changeb(b[level], b_recieve, NB);
        }
        int num = ((level + 1) * NB > N) ? num = N - level * NB : NB;
        pangulu_sptrsv_dispose_L(num,
                                 num,
                                 L_Smatrix_value,
                                 x[level],
                                 b[level],
                                 CUDA_x,
                                 CUDA_b,
                                 d_graphInDegree,
                                 d_graphInDegree_backup,
                                 d_left_sum,
                                 d_id_extractor);
        for (int i = 0; i < NB; i++)
        {
            y1[level][i] = x[level][i];
        }
        int send_id = rank;
        for (int i = 0; i < ((P - 1) < (block_length - level - 1) ? (P - 1) : (block_length - level - 1)); i++)
        {
            send_id += Q;
            if (send_id >= P * Q)
            {
                send_id = send_id % Q;
            }

            pangulu_sptrsv_send_vector(x[level], send_id, level * block_length + level, NB);
        }
    }

    void pangulu_sptrsv_dispose_block_pangulu_Smatrix_U(int rank,
                                                        int block_row,
                                                        int block_length,
                                                        int P,
                                                        int Q,
                                                        int NB,
                                                        int N,
                                                        pangulu_Smatrix *U_Smatrix_value,
                                                        calculate_type **b,
                                                        calculate_type **x,
                                                        calculate_type **b_now,
                                                        calculate_type *x_ref,
                                                        calculate_type *b_recieve,
                                                        calculate_type *CUDA_x,
                                                        calculate_type *CUDA_b,
                                                        int *d_graphInDegree,
                                                        int *d_graphInDegree_backup,
                                                        calculate_type *d_left_sum,
                                                        int *d_id_extractor)
    {
        int level = block_row;
        int recieve_id;
        int recieve_level = ((level + Q) < block_length) ? level + Q : block_length;
        for (int i = ((level + 1) < block_length) ? level + 1 : block_length; i < recieve_level; i++)
        {
            recieve_id = i % Q + level % P * ((level % P) ? Q : 0);
            pangulu_sptrsv_recv_vector(b_recieve, recieve_id, level * block_length + i + block_length * block_length, NB);
            pangulu_sptrsv_changeb(b[level], b_recieve, NB);
        }
        int num = ((level + 1) * NB > N) ? num = N - level * NB : NB;
        pangulu_sptrsv_dispose_U(num,
                                 num,
                                 U_Smatrix_value,
                                 x[level],
                                 b[level],
                                 CUDA_x,
                                 CUDA_b,
                                 d_graphInDegree,
                                 d_graphInDegree_backup,
                                 d_left_sum,
                                 d_id_extractor);
        int send_id = rank;
        for (int i = 0; i < ((P - 1) < level ? (P - 1) : level); i++)
        {
            send_id -= Q;
            if (send_id < 0)
            {
                send_id += P * Q;
            }
            pangulu_sptrsv_send_vector(x[level],
                                       send_id,
                                       level * block_length + level + block_length * block_length,
                                       NB);
        }
    }

    void pangulu_sptrsv_solve_block_pangulu_Smatrix_L(int rank,
                                                      int block_row,
                                                      int block_col,
                                                      int block_length,
                                                      int P,
                                                      int Q,
                                                      int NB,
                                                      pangulu_Smatrix *L_Smatrix_value,
                                                      calculate_type **b,
                                                      calculate_type *x,
                                                      calculate_type **b_now,
                                                      calculate_type *Left,
                                                      calculate_type *spmv_CUDA_x,
                                                      calculate_type *spmv_CUDA_y)
    {
        for (int i = 0; i < NB; i++)
        {
            Left[i] = 0.0;
        }
        if (L_Smatrix_value != NULL)
        {
            pangulu_sptrsv_spmv(NB, NB, L_Smatrix_value, x, Left, spmv_CUDA_x, spmv_CUDA_y);
        }
        int send_id = block_row % P * (block_row % P ? Q : 0) + (block_row % Q);
        if (send_id == rank)
        {
            pangulu_sptrsv_changeb(b[block_row], Left, NB);
        }
        else
        {
            pangulu_sptrsv_add_b(b_now[block_row], Left, NB);
        }
    }

    void pangulu_sptrsv_solve_block_pangulu_Smatrix_U(int rank,
                                                      int block_row,
                                                      int block_col,
                                                      int block_length,
                                                      int P,
                                                      int Q,
                                                      int NB,
                                                      pangulu_Smatrix *U_Smatrix_value,
                                                      calculate_type **b,
                                                      calculate_type *x,
                                                      calculate_type **b_now,
                                                      calculate_type *Left,
                                                      calculate_type *spmv_CUDA_x,
                                                      calculate_type *spmv_CUDA_y)
    {
        for (int i = 0; i < NB; i++)
        {
            Left[i] = 0.0;
        }
        if (U_Smatrix_value != NULL)
        {
            pangulu_sptrsv_spmv(NB, NB, U_Smatrix_value, x, Left, spmv_CUDA_x, spmv_CUDA_y);
        }
        int send_id = block_row % P * (block_row % P ? Q : 0) + (block_row % Q);
        if (send_id == rank)
        {
            pangulu_sptrsv_changeb(b[block_row], Left, NB);
        }
        else
        {
            pangulu_sptrsv_add_b(b_now[block_row], Left, NB);
        }
    }

    void pangulu_sptrsv_recv_b(calculate_type **b,
                               calculate_type *value,
                               int P,
                               int Q,
                               int rank,
                               int block_length,
                               int N,
                               int NB)
    {
        if (value == NULL)
        {
            printf("rank %d malloc error\n", rank);
            fflush(NULL);
            exit(0);
        }
        MPI_Status status;
        if (block_length != 0)
            MPI_Recv(value, block_length * NB, MPI_DOUBLE, 0, 0, MPI_COMM_WORLD, &status);
        int value_length = 0;
        for (int block_row = rank / Q; block_row < block_length; block_row += P)
        {
            for (int block_col = rank % Q; block_col < block_length; block_col += Q)
            {
                if (block_col > block_row)
                    break;
                if (block_col == block_row)
                {
                    int index = 0;
                    for (int b_row = 0; b_row < NB; b_row++)
                    {
                        b[block_row][index++] = value[value_length++];
                    }
                }
            }
        }
    }

    void pangulu_sptrsv_send_b(calculate_type *orig_b,
                               calculate_type *value,
                               int P,
                               int Q,
                               int block_length,
                               calculate_type **b,
                               int N,
                               int NB)
    {
        for (int i = 1; i < P * Q; i++)
        {
            int index_rank = i;
            int value_length = 0;
            for (int block_row = i / Q; block_row < block_length; block_row += P)
            {
                for (int block_col = i % Q; block_col < block_length; block_col += Q)
                {
                    if (block_col > block_row)
                        break;
                    int block_offset_row = block_row * NB;
                    if (block_col == block_row)
                    {
                        for (int b_row = block_offset_row; b_row < (((block_offset_row + NB) < N) ? (block_offset_row + NB) : N); b_row++)
                        {
                            value[value_length++] = orig_b[b_row];
                        }
                    }
                }
            }
            MPI_Send(value, value_length, MPI_DOUBLE, index_rank, 0, MPI_COMM_WORLD);
        }
        for (int block_row = 0; block_row < block_length; block_row += P)
        {
            for (int block_col = 0; block_col < block_length; block_col += Q)
            {
                if (block_col > block_row)
                    break;
                int block_offset_row = block_row * NB;
                if (block_col == block_row)
                {
                    int index = 0;
                    for (int b_row = block_offset_row; b_row < (((block_offset_row + NB) < N) ? (block_offset_row + NB) : N); b_row++)
                    {
                        b[block_row][index++] = orig_b[b_row];
                    }
                }
            }
        }
    }

    void pangulu_sptrsv_gather_b_send(calculate_type **b,
                                      calculate_type *value,
                                      int P,
                                      int Q,
                                      int block_length,
                                      int N,
                                      int NB,
                                      int rank)
    {
        int value_length = 0;
        for (int block_row = rank / Q; block_row < block_length; block_row += P)
        {
            for (int block_col = rank % Q; block_col < block_length; block_col += Q)
            {
                if (block_col > block_row)
                    break;
                if (block_col == block_row)
                {
                    int index = 0;
                    for (int b_row = 0; b_row < NB; b_row++)
                    {
                        value[value_length++] = b[block_row][index++];
                    }
                }
            }
        }
        MPI_Send(value, value_length, MPI_DOUBLE, 0, rank, MPI_COMM_WORLD);
    }

    void pangulu_sptrsv_gather_b_recv(calculate_type **b,
                                      calculate_type *value,
                                      int P,
                                      int Q,
                                      int block_length,
                                      int N,
                                      int NB)
    {
        if (value == NULL)
        {
            printf("rank 0 recv b malloc error\n");
            fflush(NULL);
            exit(0);
        }
        for (int i = 1; i < P * Q; i++)
        {
            MPI_Status status;
            MPI_Recv(value, block_length * NB, MPI_DOUBLE, i, i, MPI_COMM_WORLD, &status);
            int value_length = 0;
            for (int block_row = i / Q; block_row < block_length; block_row += P)
            {
                for (int block_col = i % Q; block_col < block_length; block_col += Q)
                {
                    if (block_col > block_row)
                        break;
                    if (block_col == block_row)
                    {
                        int index = 0;
                        for (int b_row = 0; b_row < NB; b_row++)
                        {
                            b[block_row][index++] = value[value_length++];
                        }
                    }
                }
            }
        }
    }

    void pangulu_sptrsv_init_b(calculate_type *orig_b,
                               calculate_type *x_ref,
                               int N,
                               calculate_type *b1,
                               calculate_type *b2)
    {
        for (int i = 0; i < N; i++)
        {
            x_ref[i] = b1[i];
            orig_b[i] = b2[i];
        }
    }

    void pangulu_sptrsv_malloc_init_sptrsv(calculate_type **b,
                                           calculate_type **y,
                                           calculate_type **b_now,
                                           calculate_type **y1,
                                           calculate_type **x,
                                           calculate_type **b1_now,
                                           calculate_type **orig_b,
                                           calculate_type **x_ref,
                                           calculate_type **value,
                                           int **reci,
                                           calculate_type **b_recieve,
                                           int block_length,
                                           int N,
                                           int NB)
    {
        *orig_b = (calculate_type *)malloc(sizeof(calculate_type) * N);
        *x_ref = (calculate_type *)malloc(sizeof(calculate_type) * N);
        *value = (calculate_type *)malloc(sizeof(calculate_type) * block_length * NB);
        *reci = (int *)malloc(sizeof(int) * block_length);
        *b_recieve = (calculate_type *)malloc(sizeof(calculate_type) * NB);
        for (int i = 0; i < block_length; i++)
        {
            b[i] = (calculate_type *)malloc(sizeof(calculate_type) * NB);
            y1[i] = (calculate_type *)malloc(sizeof(calculate_type) * NB);
        }
        for (int i = 0; i < block_length; i++)
        {
            y[i] = (calculate_type *)malloc(sizeof(calculate_type) * NB);
            for (int j = 0; j < NB; j++)
            {
                y[i][j] = 0.0;
            }
        }
        for (int i = 0; i < block_length; i++)
        {
            b_now[i] = (calculate_type *)malloc(sizeof(calculate_type) * NB);
            for (int j = 0; j < NB; j++)
            {
                b_now[i][j] = 0.0;
            }
        }

        for (int i = 0; i < block_length; i++)
        {
            x[i] = (calculate_type *)malloc(sizeof(calculate_type) * NB);
            for (int j = 0; j < NB; j++)
            {
                x[i][j] = 0.0;
            }
        }
        for (int i = 0; i < block_length; i++)
        {
            b1_now[i] = (calculate_type *)malloc(sizeof(calculate_type) * NB);
            for (int j = 0; j < NB; j++)
            {
                b1_now[i][j] = 0.0;
            }
        }
    }

    void pangulu_sptrsv_free(int block_length,
                             calculate_type **b,
                             calculate_type **y,
                             calculate_type **b_now,
                             calculate_type **y1,
                             calculate_type **x,
                             calculate_type **b1_now)
    {
        for (int i = 0; i < block_length; i++)
        {
            if (b[i] != NULL)
            {
                free(b[i]);
                b[i] = NULL;
            }
        }
        for (int i = 0; i < block_length; i++)
        {
            if (y[i] != NULL)
            {
                free(y[i]);
                y[i] = NULL;
            }
        }
        for (int i = 0; i < block_length; i++)
        {
            if (b_now[i] != NULL)
            {
                free(b_now[i]);
                b_now[i] = NULL;
            }
        }
        for (int i = 0; i < block_length; i++)
        {
            if (y1[i] != NULL)
            {
                free(y1[i]);
                y[i] = NULL;
            }
        }
        for (int i = 0; i < block_length; i++)
        {
            if (x[i] != NULL)
            {
                free(x[i]);
                x[i] = NULL;
            }
        }
        for (int i = 0; i < block_length; i++)
        {
            if (b1_now[i] != NULL)
            {
                free(b1_now[i]);
                b1_now[i] = NULL;
            }
        }
    }

    void pangulu_sptrsv_test(int N,
                             int NB,
                             int rank,
                             int P,
                             int Q,
                             int block_length,
                             pangulu_vector *Y,
                             pangulu_vector *B,
                             pangulu_vector *check_X,
                             int *L_Smatrix_rowpointer,
                             int *L_Smatrix_columnindex,
                             pangulu_Smatrix **L_Smatrix_value,
                             int *U_Smatrix_rowpointer,
                             int *U_Smatrix_columnindex,
                             pangulu_Smatrix **U_Smatrix_value)
    {
        calculate_type *b1 = Y->value;
        calculate_type *b2 = B->value;
        calculate_type *b3 = check_X->value;
        calculate_type *spmv_CUDA_x;
        calculate_type *spmv_CUDA_y;
        int *d_graphInDegree;
        int *d_graphInDegree_backup;
        calculate_type *d_left_sum;
        int *d_id_extractor;
        calculate_type *sptrsv_CUDA_x;
        calculate_type *sptrsv_CUDA_b;
        pangulu_sptrsv_cuda_malloc_init(&spmv_CUDA_x,
                                        &spmv_CUDA_y,
                                        &sptrsv_CUDA_x,
                                        &sptrsv_CUDA_b,
                                        &d_graphInDegree,
                                        &d_graphInDegree_backup,
                                        &d_left_sum,
                                        &d_id_extractor,
                                        NB);
        calculate_type *orig_b;
        calculate_type *x_ref;
        calculate_type *value;
        int *reci;
        calculate_type *b_recieve;
        calculate_type **b = (calculate_type **)malloc(sizeof(calculate_type *) * block_length);
        calculate_type **y = (calculate_type **)malloc(sizeof(calculate_type *) * block_length);
        calculate_type **b_now = (calculate_type **)malloc(sizeof(calculate_type *) * block_length);
        calculate_type **y1 = (calculate_type **)malloc(sizeof(calculate_type *) * block_length);
        calculate_type **x = (calculate_type **)malloc(sizeof(calculate_type *) * block_length);
        calculate_type **b1_now = (calculate_type **)malloc(sizeof(calculate_type *) * block_length);
        pangulu_sptrsv_malloc_init_sptrsv(b,
                                          y,
                                          b_now,
                                          y1,
                                          x,
                                          b1_now,
                                          &orig_b,
                                          &x_ref,
                                          &value,
                                          &reci,
                                          &b_recieve,
                                          block_length,
                                          N,
                                          NB);
        if (rank == 0)
        {
            pangulu_sptrsv_init_b(orig_b, x_ref, N, b1, b2);
            pangulu_sptrsv_send_b(orig_b, value, P, Q, block_length, b, N, NB);
        }
        else
        {
            pangulu_sptrsv_recv_b(b, value, P, Q, rank, block_length, N, NB);
        }
        for (int i = 0; i < block_length; i++)
            reci[i] = 0;
        for (int i = 0; i < N; i++)
            b3[i] = 2.0;
        int offcol = rank % Q;
        calculate_type check_error = 0;
        MPI_Barrier(MPI_COMM_WORLD);
        for (int i = rank / Q; i < block_length; i += P)
        {
            for (int j = L_Smatrix_rowpointer[i]; j < L_Smatrix_rowpointer[i + 1]; j++)
            {
                int block_row = i;
                int block_col = L_Smatrix_columnindex[j];
                if (block_row == block_col)
                {
                    pangulu_sptrsv_dispose_block_pangulu_Smatrix_L(rank,
                                                                   block_row,
                                                                   block_length,
                                                                   P,
                                                                   Q,
                                                                   NB,
                                                                   N,
                                                                   L_Smatrix_value[j],
                                                                   b,
                                                                   y,
                                                                   b_now,
                                                                   b1,
                                                                   b_recieve,
                                                                   y1,
                                                                   sptrsv_CUDA_x,
                                                                   sptrsv_CUDA_b,
                                                                   d_graphInDegree,
                                                                   d_graphInDegree_backup,
                                                                   d_left_sum,
                                                                   d_id_extractor);
                    reci[block_col] = 1;
                }
                else
                {
                    int offset = block_col % P * (block_col % P ? Q : 0) + (block_col % Q);
                    if (rank != offset && reci[block_col] == 0)
                    {

                        pangulu_sptrsv_recv_vector(y[block_col],
                                                   offset,
                                                   block_col * block_length + block_col,
                                                   NB);
                        reci[block_col] = 1;
                    }
                    pangulu_sptrsv_solve_block_pangulu_Smatrix_L(rank,
                                                                 block_row,
                                                                 block_col,
                                                                 block_length,
                                                                 P,
                                                                 Q,
                                                                 NB,
                                                                 L_Smatrix_value[j],
                                                                 b,
                                                                 y[block_col],
                                                                 b_now,
                                                                 b_recieve,
                                                                 spmv_CUDA_x,
                                                                 spmv_CUDA_y);
                }
            }
            int last_col = (i / Q) * Q + offcol;
            if (last_col != i)
            {
                int send_id = i % P * (i % P ? Q : 0) + (i % Q);
                if (last_col > i)
                    last_col -= Q;
                int index_value = i * block_length + last_col;

                if (last_col >= 0)
                {
                    pangulu_sptrsv_send_vector(b_now[i], send_id, index_value, NB);
                }
            }
        }
        for (int i = 0; i < block_length; i++)
            reci[i] = 0;
        int first_level = rank / Q + ((block_length - 1) / P) * P;
        first_level = first_level < block_length ? first_level : first_level - P;
        for (int i = first_level; i >= 0; i -= P)
        {
            for (int j = U_Smatrix_rowpointer[i + 1] - 1; j >= U_Smatrix_rowpointer[i]; j--)
            {
                int block_row = i;
                int block_col = U_Smatrix_columnindex[j];
                if (block_row == block_col)
                {
                    pangulu_sptrsv_dispose_block_pangulu_Smatrix_U(rank,
                                                                   block_row,
                                                                   block_length,
                                                                   P,
                                                                   Q,
                                                                   NB,
                                                                   N,
                                                                   U_Smatrix_value[j],
                                                                   y1,
                                                                   x,
                                                                   b1_now,
                                                                   b3,
                                                                   b_recieve,
                                                                   sptrsv_CUDA_x,
                                                                   sptrsv_CUDA_b,
                                                                   d_graphInDegree,
                                                                   d_graphInDegree_backup,
                                                                   d_left_sum,
                                                                   d_id_extractor);
                    reci[block_col] = 1;
                }
                else
                {
                    int offset = block_col % P * (block_col % P ? Q : 0) + (block_col % Q);
                    if (rank != offset && reci[block_col] == 0)
                    {
                        pangulu_sptrsv_recv_vector(x[block_col],
                                                   offset,
                                                   block_col * block_length + block_col + block_length * block_length,
                                                   NB);
                        reci[block_col] = 1;
                    }
                    pangulu_sptrsv_solve_block_pangulu_Smatrix_U(rank,
                                                                 block_row,
                                                                 block_col,
                                                                 block_length,
                                                                 P,
                                                                 Q,
                                                                 NB,
                                                                 U_Smatrix_value[j],
                                                                 y1,
                                                                 x[block_col],
                                                                 b1_now,
                                                                 b_recieve,
                                                                 spmv_CUDA_x,
                                                                 spmv_CUDA_y);
                }
            }
            int last_col = (i / Q) * Q + offcol;
            if (last_col != i)
            {
                int send_id = i % P * (i % P ? Q : 0) + (i % Q);
                if (last_col < i)
                    last_col += Q;
                int index_value = i * block_length + last_col;

                if (last_col < block_length)
                {
                    pangulu_sptrsv_send_vector(b1_now[i],
                                               send_id, index_value + block_length * block_length, NB);
                }
            }
        }
        if (rank == 0)
        {
            pangulu_sptrsv_gather_b_recv(x, value, P, Q, block_length, N, NB);
        }
        else
        {
            pangulu_sptrsv_gather_b_send(x, value, P, Q, block_length, N, NB, rank);
        }
        if (rank == 0)
        {
            check_error = pangulu_sptrsv_check_calculate_x(b3, x, block_length, NB, N);
            if (check_error > CHECK_ERROR || check_error < -CHECK_ERROR)
            {
                //printf("SpTRSV check_error = %lf\n", check_error);
                printf("CHECK------------------------FAILED!!!\n");
            }
            else
            {
                printf("CHECK------------------------PASSED!!!\n");
            }
        }
        free(orig_b);
        free(x_ref);
        free(value);
        free(reci);
        free(b_recieve);
        pangulu_sptrsv_free(block_length, b, y, b_now, y1, x, b1_now);
        free(b);
        free(y);
        free(b_now);
        free(y1);
        free(x);
        free(b1_now);
        pangulu_sptrsv_cuda_malloc_free(&spmv_CUDA_x,
                                        &spmv_CUDA_y,
                                        &sptrsv_CUDA_x,
                                        &sptrsv_CUDA_b,
                                        &d_graphInDegree,
                                        &d_graphInDegree_backup,
                                        &d_left_sum,
                                        &d_id_extractor,
                                        NB);
    }
    void pangulu_sptrsv(int N,
                        int rank,
                        int P,
                        int Q,
                        int NB,
                        int block_length,
                        pangulu_Smatrix *symbolic_Smatrix,
                        int *L_pangulu_Smatrix_columnpointer,
                        int *L_pangulu_Smatrix_rowindex,
                        pangulu_Smatrix **L_pangulu_Smatrix_value,
                        int *U_pangulu_Smatrix_rowpointer,
                        int *U_pangulu_Smatrix_columnindex,
                        pangulu_Smatrix **U_pangulu_Smatrix_value)
    {
        int *L_pangulu_Smatrix_rowpointer;
        int *L_pangulu_Smatrix_columnindex;
        pangulu_Smatrix **L_pangulu_Smatrix_value_CSR;
        int vector_length = ((N + NB - 1) / NB) * NB;
        pangulu_vector *X = (pangulu_vector *)malloc(sizeof(pangulu_vector));
        pangulu_read_pangulu_vector(X, vector_length);
        pangulu_vector *Y = (pangulu_vector *)malloc(sizeof(pangulu_vector));
        pangulu_init_pangulu_vector(Y, vector_length);
        pangulu_vector *B = (pangulu_vector *)malloc(sizeof(pangulu_vector));
        pangulu_init_pangulu_vector(B, vector_length);
        pangulu_pangulu_Smatrix_multiple_pangulu_vector(symbolic_Smatrix, X, B);
        pangulu_block_pangulu_Smatrix_CSC_to_block_pangulu_Smatrix_CSR(rank,
                                                                       block_length,
                                                                       P,
                                                                       Q,
                                                                       L_pangulu_Smatrix_columnpointer,
                                                                       L_pangulu_Smatrix_rowindex,
                                                                       L_pangulu_Smatrix_value,
                                                                       &L_pangulu_Smatrix_rowpointer,
                                                                       &L_pangulu_Smatrix_columnindex,
                                                                       &L_pangulu_Smatrix_value_CSR);
        pangulu_sptrsv_test(N,
                            NB,
                            rank,
                            P, Q,
                            block_length,
                            Y,
                            B,
                            X,
                            L_pangulu_Smatrix_rowpointer,
                            L_pangulu_Smatrix_columnindex,
                            L_pangulu_Smatrix_value_CSR,
                            U_pangulu_Smatrix_rowpointer,
                            U_pangulu_Smatrix_columnindex,
                            U_pangulu_Smatrix_value);
        X = pangulu_destroy_pangulu_vector(X);
        Y = pangulu_destroy_pangulu_vector(Y);
        B = pangulu_destroy_pangulu_vector(B);
    }

#ifdef __cplusplus
}
#endif

#endif