#ifndef PANGULU_SPTRSV_CUDA_H
#define PANGULU_SPTRSV_CUDA_H
#ifdef __cplusplus
extern "C"
{
#endif
#include "pangulu_cuda_interface.h"
    /*
The function in this file is primarily called kernel,these functions would be called in pangulu_sptrsv.h
*/

    void pangulu_sptrsv_spmv_inteface(pangulu_Smatrix *A,
                                      int row,
                                      double *x,
                                      double *y,
                                      double *CUDA_x,
                                      double *CUDA_y)
    {
        if (A->CUDA_rowpointer == NULL)
        {
            return;
        }
        int n = A->row;
        pangulu_cuda_memcpy_value_host_to_device(A);
        pangulu_cuda_memcpy_host_to_device_double(CUDA_x, x, row);
        pangulu_sptrsv_spmv_kernel(n, A->CUDA_rowpointer, A->CUDA_columnindex, A->CUDA_value, CUDA_x, CUDA_y);
        pangulu_cuda_memcpy_device_to_host_double(y, CUDA_y, row);
    }

    void pangulu_sptrsv_dispose_L_interface(pangulu_Smatrix *A,
                                            int row,
                                            int colnum,
                                            double *x,
                                            double *b,
                                            double *CUDA_x,
                                            double *CUDA_b,
                                            int *d_graphInDegree,
                                            int *d_graphInDegree_backup,
                                            double *d_left_sum,
                                            int *d_id_extractor)
    {
        pangulu_cuda_memcpy_value_host_to_device(A);
        pangulu_cuda_transport_CSR_to_CSC(A);
        pangulu_cuda_memcpy_host_to_device_double(CUDA_b, b, row);
        int nnz = A->rowpointer[row];
        pangulu_sptrsv_dispose_L_kernel(row,
                                        colnum,
                                        nnz,
                                        A->CUDA_columnpointer,
                                        A->CUDA_rowindex,
                                        A->CUDA_value_CSC,
                                        CUDA_x,
                                        CUDA_b,
                                        d_graphInDegree,
                                        d_graphInDegree_backup,
                                        d_left_sum,
                                        d_id_extractor);
        pangulu_cuda_memcpy_device_to_host_double(x, CUDA_x, row);
    }

    void pangulu_sptrsv_dispose_U_interface(pangulu_Smatrix *A,
                                            int row,
                                            int colnum,
                                            double *x,
                                            double *b,
                                            double *CUDA_x,
                                            double *CUDA_b,
                                            int *d_graphInDegree,
                                            int *d_graphInDegree_backup,
                                            double *d_left_sum,
                                            int *d_id_extractor)
    {
        pangulu_cuda_memcpy_value_host_to_device(A);
        pangulu_cuda_transport_CSR_to_CSC(A);
        pangulu_cuda_memcpy_host_to_device_double(CUDA_b, b, row);
        int nnz = A->rowpointer[row];
        pangulu_sptrsv_dispose_U_kernel(row,
                                        colnum,
                                        nnz,
                                        A->CUDA_columnpointer,
                                        A->CUDA_rowindex,
                                        A->CUDA_value_CSC,
                                        CUDA_x,
                                        CUDA_b,
                                        d_graphInDegree,
                                        d_graphInDegree_backup,
                                        d_left_sum,
                                        d_id_extractor);
        pangulu_cuda_memcpy_device_to_host_double(x, CUDA_x, row);
    }

#ifdef __cplusplus
}
#endif

#endif