#ifndef PANGULU_SSSSM_FP64_H
#define PANGULU_SSSSM_FP64_H
#ifdef __cplusplus
extern "C"
{
#endif
#include "pangulu_common.h"
    /*
The function in this file is primarily called kernel,these functions would be called in pangulu_kernel_interface.h
*/
    const int *lower_bound(const int *begins, const int *ends, int key)
    {
        int curSize = 0;
        curSize = ends - begins;
        int half;
        const int *middle;
        while (curSize > 0)
        {
            half = curSize >> 1;
            middle = begins + half;
            if (*middle < key)
            {
                begins = middle;
                ++begins;
                curSize = curSize - half - 1;
            }
            else
                curSize = half;
        }
        return begins;
    }

    void SpMM(
        int m, int k, int n,
        const int *A_csrOffsets,
        const int *A_columns,
        const double *A_values,
        const int *B_csrOffsets,
        const int *B_columns,
        const double *B_values,
        const int *C_csrOffsets,
        const int *C_columns,
        double *C_values)
    {
        for (int rowA = 0; rowA < m; ++rowA)
        {
            for (int i = A_csrOffsets[rowA]; i < A_csrOffsets[rowA + 1]; ++i)
            {
                int rowB = A_columns[i];
                for (int j = B_csrOffsets[rowB]; j < B_csrOffsets[rowB + 1]; ++j)
                {
                    int colB = B_columns[j];
                    ///find locate of C at [rowA,colB]
                    int loc = (int)(lower_bound(
                                        C_columns + C_csrOffsets[rowA],
                                        C_columns + C_csrOffsets[rowA + 1],
                                        colB) -
                                    C_columns);
                    if (C_columns[loc] == colB)
                    {
                        //printf("ERROR\n");
                        C_values[loc] -= A_values[i] * B_values[j];
                    }
                }
            }
        }
    }
    void pangulu_ssssm_fp64(pangulu_Smatrix *A, pangulu_Smatrix *L, pangulu_Smatrix *U)
    {
        int n = A->row;
        SpMM(n, n, n, L->rowpointer, L->columnindex, L->value, U->rowpointer, U->columnindex, U->value, A->rowpointer, A->columnindex, A->value);
    }
    
#ifdef __cplusplus
}
#endif

#endif