#ifndef PANGULU_UTILS_H
#define PANGULU_UTILS_H
#ifdef __cplusplus
extern "C"
{
#endif
#include "mmio.h"
#include "mmio_highlevel.h"
#include "pangulu_common.h"
#include <mpi.h>
#include <string.h>
#include "pangulu_cuda_interface.h"
#include <getopt.h>
    /*
The function in this file is primarily called kernel,these functions would be called in many other head files
*/
    void pangulu_get_common(pangulu_common *common,
                            int ARGC, char **ARGV)
    {
        int c;
        extern char *optarg;
        common->NB = 0;
        common->P = 0;
        common->Q = 0;
        while ((c = getopt(ARGC, ARGV, "P:Q:NB:F:")) != EOF)
        {
            switch (c)
            {
            case 'P':
                common->P = atoi(optarg);
                continue;
            case 'Q':
                common->Q = atoi(optarg);
                continue;
            case 'B':
                common->NB = atoi(optarg);
                continue;
            case 'F':
                common->file_name = (char *)malloc((strlen(optarg)) * sizeof(char));
                strcpy(common->file_name, optarg);
                continue;
            }
        }
        if (common->rank == 0)
        {
            printf("rank is %d P is %d Q is %d NB is %d\n", common->rank, common->P, common->Q, common->NB);
        }
        if ((common->P * common->Q) != common->size)
        {
            printf("error P * Q != size\n");
            exit(3);
        }
        if ((common->NB) == 0)
        {
            printf("error the NB is 0\n");
            exit(4);
        }
    }

    void pangulu_malloc_pangulu_Smatrix(pangulu_Smatrix *S,
                                        int NB, int *save_rowpointer)
    {
        int *rowpointer = (int *)malloc(sizeof(int) * (NB + 1));
        S->row = NB;
        S->column = NB;
        for (int i = 0; i < NB; i++)
        {
            rowpointer[i + 1] = save_rowpointer[i];
        }
        rowpointer[0] = 0;
        int nnzA = rowpointer[NB];
        int *columindex = (int *)malloc(sizeof(int) * nnzA);
        calculate_type *value = (calculate_type *)malloc(sizeof(calculate_type) * nnzA);
        for (int i = 0; i < nnzA; i++)
        {
            columindex[i] = 0;
        }

        for (int i = 0; i < nnzA; i++)
        {
            value[i] = 0.0;
        }
        S->rowpointer = rowpointer;
        S->columnindex = columindex;
        S->value = value;
    }

    void pangulu_send_vector_int(int *A, int N, int send_id, int signal)
    {
        MPI_Send(A, N, MPI_INT, send_id, signal, MPI_COMM_WORLD);
    }

    void pangulu_recv_vector_int(int *A, int N, int receive_id, int signal)
    {
        MPI_Status status;
        for (int i = 0; i < N; i++)
        {
            A[i] = 0;
        }
        MPI_Recv(A, N, MPI_INT, receive_id, signal, MPI_COMM_WORLD, &status);
    }

    void pangulu_send_vector_double(double *A, int N, int send_id, int signal)
    {
        MPI_Send(A, N, MPI_DOUBLE, send_id, signal, MPI_COMM_WORLD);
    }

    void pangulu_recv_vector_double(double *A, int N, int receive_id, int signal)
    {
        MPI_Status status;
        for (int i = 0; i < N; i++)
        {
            A[i] = 0.0;
        }
        MPI_Recv(A, N, MPI_DOUBLE, receive_id, signal, MPI_COMM_WORLD, &status);
    }

    void pangulu_Bcast_vector(int *vector, int length, int send_rank)
    {
        MPI_Bcast(vector, length, MPI_INT, send_rank, MPI_COMM_WORLD);
    }

    void pangulu_time_start(pangulu_common *common)
    {
        gettimeofday(&(common->start_time), NULL);
    }
    void pangulu_time_stop(pangulu_common *common)
    {
        gettimeofday(&(common->stop_time), NULL);
    }
    double pangulu_get_spend_time(pangulu_common *common)
    {
        double time = (common->stop_time.tv_sec - common->start_time.tv_sec) * 1000.0 + (common->stop_time.tv_usec - common->start_time.tv_usec) / 1000.0;
        return time / 1000.0;
    }

    int n_factorial(int n)
    {
        if (n < 2)
        {
            return 1;
        }
        else
        {
            return (n * (n + 1)) >> 1;
        }
    }

    int pangulu_Bcast_pangulu_Smatrix(pangulu_Smatrix *S, int rank)
    {

        int n = S->row;
        int nnz;
        if (rank == 0)
        {
            MPI_Bcast(&n, 1, MPI_INT, 0, MPI_COMM_WORLD);
            MPI_Bcast(S->rowpointer, n + 1, MPI_INT, 0, MPI_COMM_WORLD);
            nnz = S->rowpointer[n];
            MPI_Bcast(S->columnindex, nnz, MPI_INT, 0, MPI_COMM_WORLD);
            MPI_Bcast(S->value, nnz, MPI_DOUBLE, 0, MPI_COMM_WORLD);
        }
        else
        {
            MPI_Bcast(&n, 1, MPI_INT, 0, MPI_COMM_WORLD);
            S->rowpointer = (int *)malloc(sizeof(int) * (n + 1));
            MPI_Bcast(S->rowpointer, n + 1, MPI_INT, 0, MPI_COMM_WORLD);
            nnz = S->rowpointer[n];
            S->columnindex = (int *)malloc(sizeof(int) * nnz);
            S->value = (calculate_type *)malloc(sizeof(calculate_type) * nnz);
            S->row = n;
            S->column = n;
            MPI_Bcast(S->columnindex, nnz, MPI_INT, 0, MPI_COMM_WORLD);
            MPI_Bcast(S->value, nnz, MPI_DOUBLE, 0, MPI_COMM_WORLD);
        }
        return n;
    }

    void pangulu_send_pangulu_Smatrix(pangulu_Smatrix *S,
                                      int send_id, int signal, int N)
    {
        MPI_Send(&N, 1, MPI_INT, send_id, signal, MPI_COMM_WORLD);
        MPI_Send(S->rowpointer, N + 1, MPI_INT, send_id, signal, MPI_COMM_WORLD);
        MPI_Send(S->columnindex, S->rowpointer[N], MPI_INT, send_id, signal, MPI_COMM_WORLD);
        MPI_Send(S->value, S->rowpointer[N], MPI_DOUBLE, send_id, signal, MPI_COMM_WORLD);
    }

    void pangulu_recv_pangulu_Smatrix(pangulu_Smatrix *S, int receive_id, int signal, int *row)
    {
        MPI_Status status;
        MPI_Recv(row, 1, MPI_INT, receive_id, signal, MPI_COMM_WORLD, &status);
        S->row = *row;
        int N = *row;
        S->rowpointer = (int *)malloc(sizeof(int) * (N + 1));
        for (int i = 0; i < (N + 1); i++)
        {
            S->rowpointer[i] = 0;
        }
        MPI_Recv(S->rowpointer, N + 1, MPI_INT, receive_id, signal, MPI_COMM_WORLD, &status);
        int nnz = S->rowpointer[N];
        S->columnindex = (int *)malloc(sizeof(int) * nnz);
        S->value = (calculate_type *)malloc(sizeof(calculate_type) * nnz);
        for (int i = 0; i < S->rowpointer[N]; i++)
        {
            S->columnindex[i] = 0;
        }
        for (int i = 0; i < S->rowpointer[N]; i++)
        {
            S->value[i] = 0.0;
        }
        MPI_Recv(S->columnindex, S->rowpointer[N], MPI_INT, receive_id, signal, MPI_COMM_WORLD, &status);
        MPI_Recv(S->value, S->rowpointer[N], MPI_DOUBLE, receive_id, signal, MPI_COMM_WORLD, &status);
        S->row = N;
        S->column = N;
    }

    int pangulu_Bcast_N(int N, int send_rank)
    {
        MPI_Bcast(&N, 1, MPI_INT, send_rank, MPI_COMM_WORLD);
        return N;
    }

    void pangulu_read_common(pangulu_common *common)
    {
        FILE *fp = fopen("./text.txt", "r");
        if (fp == NULL)
        {
            printf("open file name = text.txt error please check this file\n");
            exit(0);
        }
        char *annotation;
        annotation = (char *)malloc(sizeof(char) * 100);
        int flag;
        flag = fscanf(fp, "%d %s", &(common->NB), annotation);
        flag = fscanf(fp, "%d %s", &(common->P), annotation);
        flag = fscanf(fp, "%d %s", &(common->Q), annotation);
        free(annotation);
        fclose(fp);
        return;
    }

    void pangulu_read_pangulu_Smatrix(pangulu_Smatrix *S,
                                      char *filename)
    {
        int isSymmeticeR;
        int nnz;
        int nrow, ncol;
        mmio_info(&nrow, &ncol, &nnz, &isSymmeticeR, filename);
        int *rowpointer = (int *)malloc(sizeof(int) * (nrow + 1));
        int *columnindex = (int *)malloc(sizeof(int) * nnz);
        float *save_value = (float *)malloc(sizeof(float) * nnz);
        mmio_data(rowpointer, columnindex, save_value, filename);
        S->row = nrow;
        S->column = ncol;
        S->rowpointer = rowpointer;
        S->columnindex = columnindex;
        calculate_type *value = (calculate_type *)malloc(sizeof(calculate_type) * nnz);
        for (int i = 0; i < nnz; i++)
        {
            value[i] = save_value[i];
        }
        free(save_value);
        S->value = value;
    }

    void pangulu_read_pangulu_vector(pangulu_vector *X, int n)
    {
        X->value = (calculate_type *)malloc(sizeof(calculate_type) * n);
        for (int i = 0; i < n; i++)
        {
            X->value[i] = 2.0;
        }
        X->row = n;
    }
    void pangulu_init_pangulu_vector(pangulu_vector *B, int n)
    {
        B->value = (calculate_type *)malloc(sizeof(calculate_type) * n);
        for (int i = 0; i < n; i++)
        {
            B->value[i] = 0.0;
        }
        B->row = n;
    }
    void pangulu_pangulu_Smatrix_multiple_pangulu_vector(pangulu_Smatrix *A,
                                                         pangulu_vector *X,
                                                         pangulu_vector *B)
    {
        int n = A->row;
        calculate_type *X_value = X->value;
        calculate_type *B_value = B->value;
        for (int i = 0; i < n; i++)
        {
            for (int j = A->rowpointer[i]; j < A->rowpointer[i + 1]; j++)
            {
                int col = A->columnindex[j];
                B_value[i] += A->value[j] * X_value[col];
            }
        }
    }

    void pangulu_Check_pangulu_Smatrix(pangulu_Smatrix *B)
    {
        printf("------------------\n\n\n");
        if (B == NULL)
        {
            printf("\nno i am null\n");
            return;
        }
        printf("row is %d column is %d\n", B->row, B->column);
        printf("rowpointer:");
        for (int i = 0; i < B->row + 1; i++)
        {
            printf("%d ", B->rowpointer[i]);
        }
        printf("\n");
        printf("columnindex:\n");
        for (int i = 0; i < B->rowpointer[B->row]; i++)
        {
            printf("%d ", B->columnindex[i]);
        }
        printf("\n");
        printf("value:\n");
        for (int i = 0; i < B->rowpointer[B->row]; i++)
        {
            printf("%lf ", B->value[i]);
        }
        printf("\n\n\n--------------------");
    }

    void pangulu_init_pangulu_Smatrix(pangulu_Smatrix *S)
    {
        S->value = NULL;
        S->value_CSC = NULL;

        S->CSR_to_CSC_index = NULL;
        S->CSC_to_CSR_index = NULL;

        S->rowpointer = NULL;
        S->columnindex = NULL;
        S->columnpointer = NULL;
        S->rowindex = NULL;
        S->column = 0;
        S->row = 0;

        S->bin_rowpointer = NULL;
        S->bin_rowindex = NULL;
        S->CUDA_nnzU = NULL;
        S->nnzU = NULL;
        S->CUDA_value = NULL;
        S->CUDA_value_CSC = NULL;
        S->CUDA_CSR_to_CSC_index = NULL;
        S->CUDA_CSC_to_CSR_index = NULL;
        S->CUDA_rowpointer = NULL;
        S->CUDA_columnindex = NULL;
        S->CUDA_columnpointer = NULL;
        S->CUDA_rowindex = NULL;
    }

    int pangulu_bin_map(int nnz)
    {
        int log_nnz = ceil(log2((double)nnz));
        if (log_nnz >= 14)
        {
            return 11;
        }
        else if (log_nnz > 5)
        {
            return log_nnz - 2;
        }
        else if (nnz > 1)
        {
            return 3;
        }
        else if (nnz == 1)
        {
            return 2;
        }
        else
        {
            return 1;
        }
    }

    void pangulu_pangulu_Smatrix_init_malloc(pangulu_Smatrix *A)
    {
        int nnzA = A->rowpointer[A->row];
        int n = A->row;
        int *rowpointer = (int *)malloc(sizeof(int) * (n + 1));
        int *columnindex = (int *)malloc(sizeof(int) * nnzA);
        calculate_type *value = (calculate_type *)malloc(sizeof(calculate_type) * nnzA);
        int *CSR_to_CSC_index = (int *)malloc(sizeof(int) * nnzA);
        int *CSC_to_CSR_index = (int *)malloc(sizeof(int) * nnzA);
        for (int i = 0; i < nnzA; i++)
        {
            value[i] = 0.0;
        }

        for (int i = 0; i < (n + 1); i++)
        {
            rowpointer[i] = 0;
        }
        for (int i = 0; i < nnzA; i++)
        {
            rowpointer[A->columnindex[i] + 1]++;
        }
        for (int i = 0; i < n; i++)
        {
            rowpointer[i + 1] += rowpointer[i];
        }
        int *index_rowpointer = (int *)malloc(sizeof(int) * (n + 1));
        for (int i = 0; i < n; i++)
        {
            index_rowpointer[i] = rowpointer[i];
        }
        for (int i = 0; i < n; i++)
        {
            for (int j = A->rowpointer[i]; j < A->rowpointer[i + 1]; j++)
            {

                int col = A->columnindex[j];
                int index = index_rowpointer[col];
                columnindex[index] = i;
                value[index] = A->value[j];
                CSR_to_CSC_index[j] = index;
                CSC_to_CSR_index[index] = j;
                index_rowpointer[col]++;
            }
        }
        int *bin_rowpointer = (int *)malloc(sizeof(int) * BIN_LENGTH);
        int *bin_rowindex = (int *)malloc(sizeof(int) * n);
        for (int i = 0; i < BIN_LENGTH; i++)
        {
            bin_rowpointer[i] = 0;
        }
        for (int i = 0; i < n; i++)
        {
            int row_length = A->rowpointer[i + 1] - A->rowpointer[i];
            bin_rowpointer[pangulu_bin_map(row_length)]++;
        }
        int *save_bin_rowpointer = (int *)malloc(sizeof(int) * (BIN_LENGTH + 1));
        for (int i = 0; i < BIN_LENGTH - 1; i++)
        {
            bin_rowpointer[i + 1] += bin_rowpointer[i];
        }
        for (int i = 0; i < BIN_LENGTH; i++)
        {
            save_bin_rowpointer[i + 1] = bin_rowpointer[i];
        }
        for (int i = 0; i < n; i++)
        {
            int row_length = A->rowpointer[i + 1] - A->rowpointer[i];
            int index = pangulu_bin_map(row_length);
            bin_rowindex[save_bin_rowpointer[index]] = i;
            save_bin_rowpointer[index]++;
        }
        A->bin_rowpointer = bin_rowpointer;
        A->bin_rowindex = bin_rowindex;
        A->columnpointer = rowpointer;
        A->rowindex = columnindex;
        A->value_CSC = value;
        A->CSR_to_CSC_index = CSR_to_CSC_index;
        A->CSC_to_CSR_index = CSC_to_CSR_index;
        free(index_rowpointer);
        free(save_bin_rowpointer);
        pangulu_cuda_malloc_init_pangulu_Smatrix(A);
        pangulu_cuda_memcpy_init_pangulu_Smatrix(A);
    }
    void pangulu_copy_pangulu_Smatrix_to_pangulu_Smatrix(pangulu_Smatrix *S,
                                                         pangulu_Smatrix *copy_S,
                                                         int NB)
    {
        int *rowpointer = (int *)malloc(sizeof(int) * (NB + 1));
        int *columnindex = (int *)malloc(sizeof(int) * S->rowpointer[NB]);
        calculate_type *value = (calculate_type *)malloc(sizeof(calculate_type) * S->rowpointer[NB]);
        for (int i = 0; i < (NB + 1); i++)
        {
            rowpointer[i] = S->rowpointer[i];
        }
        for (int i = 0; i < S->rowpointer[NB]; i++)
        {
            columnindex[i] = S->columnindex[i];
        }
        for (int i = 0; i < S->rowpointer[NB]; i++)
        {
            value[i] = S->value[i];
        }
        copy_S->rowpointer = rowpointer;
        copy_S->columnindex = columnindex;
        copy_S->value = value;
        copy_S->row = NB;
        copy_S->column = NB;
        return;
    }

    int pangulu_mapper_to_Big_pangulu_Smatrix(int row, int col, int *mapper_A, int col_length, int P, int Q)
    {
        int mapper_index = mapper_A[(row / P) * col_length + (col / Q)];
        return mapper_index;
    }

    void pangulu_display_pangulu_Smatrix(pangulu_Smatrix *S)
    {
        printf("------------------\n\n\n");
        if (S == NULL)
        {
            printf("\nno i am null\n");
            return;
        }
        printf("row is %d column is %d\n", S->row, S->column);
        printf("rowpointer:");
        for (int i = 0; i < S->row + 1; i++)
        {
            printf("%d ", S->rowpointer[i]);
        }
        printf("\n");
        printf("columnindex:\n");
        for (int i = 0; i < S->row; i++)
        {
            for (int j = S->rowpointer[i]; j < S->rowpointer[i + 1]; j++)
            {
                printf("%d ", S->columnindex[j]);
            }
            printf("\n");
        }
        printf("\n");
        printf("value:\n");
        for (int i = 0; i < S->row; i++)
        {
            for (int j = S->rowpointer[i]; j < S->rowpointer[i + 1]; j++)
            {
                printf("%lf ", S->value[j]);
            }
            printf("\n");
        }
        printf("\n\n\n--------------------");
    }

    void pangulu_display_Big_pangulu_Smatrix(int *B_rowpointer,
                                             int *B_columnindex,
                                             pangulu_Smatrix **B_value,
                                             int rank_row_length,
                                             int rank_col_length)
    {
        printf("------------------\n\n\n");
        if (B_rowpointer == NULL)
        {
            printf("\nno i am null\n");
            return;
        }
        printf("Big Matrix row is %d column is %d\n", rank_row_length, rank_col_length);
        printf("BIG matrix rowpointer:");
        for (int i = 0; i < rank_row_length + 1; i++)
        {
            printf("%d ", B_rowpointer[i]);
        }
        printf("\n");
        printf("columnindex:\n");
        for (int i = 0; i < B_rowpointer[rank_row_length]; i++)
        {
            printf("%d ", B_columnindex[i]);
        }
        printf("\n");
        printf("value:\n");
        for (int i = 0; i < (rank_row_length + 1); i++)
        {
            for (int j = B_rowpointer[i]; j < B_rowpointer[i + 1]; j++)
            {
                printf("\nrow is %d column is %d index is %d\n", i, B_columnindex[j], j);
                pangulu_display_pangulu_Smatrix(B_value[j]);
            }
        }
        printf("\n\n\n--------------------");
    }

    void pangulu_send_pangulu_Smatrix_value(pangulu_Smatrix *S,
                                            int send_id, int signal, int NB)
    {
        MPI_Send(S->value, S->rowpointer[NB], MPI_DOUBLE, send_id, signal, MPI_COMM_WORLD);
    }

    void pangulu_recv_pangulu_Smatrix_value(pangulu_Smatrix *S,
                                            int receive_id, int signal, int NB)
    {
        MPI_Status status;
        for (int i = 0; i < S->rowpointer[NB]; i++)
        {
            S->value[i] = 0.0;
        }

        MPI_Recv(S->value, S->rowpointer[NB], MPI_DOUBLE, receive_id, signal, MPI_COMM_WORLD, &status);

        pangulu_cuda_memcpy_value_host_to_device(S);
    }

    void pangulu_send_pangulu_Smatrix_columnindex(pangulu_Smatrix *S,
                                                  int send_id, int signal, int NB)
    {
        MPI_Send(S->columnindex, S->rowpointer[NB], MPI_INT, send_id, signal, MPI_COMM_WORLD);
    }

    void pangulu_recv_pangulu_Smatrix_columnindex(pangulu_Smatrix *S,
                                                  int receive_id, int signal, int NB)
    {
        MPI_Status status;

        for (int i = 0; i < S->rowpointer[NB]; i++)
        {
            S->columnindex[i] = 0.0;
        }

        MPI_Recv(S->columnindex, S->rowpointer[NB], MPI_INT, receive_id, signal, MPI_COMM_WORLD, &status);
    }

    void pangulu_transport_pangulu_Smatrix_CSC_to_CSR(pangulu_Smatrix *S)
    {
        for (int i = 0; i < S->columnpointer[S->row]; i++)
        {
            int index = S->CSC_to_CSR_index[i];
            S->value[index] = S->value_CSC[i];
        }
    }
    void pangulu_transport_pangulu_Smatrix_CSR_to_CSC(pangulu_Smatrix *S)
    {
        for (int i = 0; i < S->rowpointer[S->row]; i++)
        {
            int index = S->CSR_to_CSC_index[i];
            S->value_CSC[index] = S->value[i];
        }
    }

    void pangulu_pangulu_Smatrix_to_dense_matrix(calculate_type *dense_matrix,
                                                 pangulu_Smatrix *S,
                                                 int N)
    {
        for (int i = 0; i < N * N; i++)
        {
            dense_matrix[i] = 0.0;
        }
        for (int i = 0; i < N; i++)
        {
            for (int j = S->rowpointer[i]; j < S->rowpointer[i + 1]; j++)
            {
                dense_matrix[i * N + S->columnindex[j]] = S->value[j];
            }
        }
    }

    void pangulu_dense_matrix_to_pangulu_Smatrix(pangulu_Smatrix *S,
                                                 calculate_type *dense_matrix,
                                                 int N)
    {

        int *rowpointer = S->rowpointer;
        int *columnindex = S->columnindex;
        calculate_type *value = S->value;
        for (int i = 0; i < N; i++)
        {
            for (int j = rowpointer[i]; j < rowpointer[i + 1]; j++)
            {
                value[j] = dense_matrix[i * N + columnindex[j]];
                if (i == columnindex[j] && value[j] < ERROR && value[j] > -ERROR)
                {
                    value[j] = ERROR;
                }
            }
        }
    }

    void pangulu_block_pangulu_Smatrix_multiply_block_pangulu_vector(pangulu_Smatrix *A,
                                                                     calculate_type *X,
                                                                     calculate_type *B)
    {
        int n = A->row;
        for (int i = 0; i < n; i++)
        {
            for (int j = A->rowpointer[i]; j < A->rowpointer[i + 1]; j++)
            {
                int col = A->columnindex[j];
                B[i] += A->value[j] * X[col];
            }
        }
    }

    void pangulu_send_pangulu_vector_value(pangulu_vector *S,
                                           int send_id, int signal, int vector_length)
    {
        MPI_Send(S->value, vector_length, MPI_DOUBLE, send_id, signal, MPI_COMM_WORLD);
    }

    void pangulu_recv_pangulu_vector_value(pangulu_vector *S, int receive_id, int signal, int vector_length)
    {
        MPI_Status status;
        MPI_Recv(S->value, vector_length, MPI_DOUBLE, receive_id, signal, MPI_COMM_WORLD, &status);
    }

    pangulu_Smatrix *pangulu_destroy_part_pangulu_Smatrix(pangulu_Smatrix *S)
    {
        free(S->rowpointer);
        free(S->columnindex);
        free(S->value);
        S->rowpointer = NULL;
        S->columnindex = NULL;
        S->value = NULL;
        free(S);
        return NULL;
    }

    pangulu_Smatrix *pangulu_destroy_pangulu_Smatrix(pangulu_Smatrix *S)
    {
        free(S->rowpointer);
        free(S->columnindex);
        free(S->value);
        free(S->CSR_to_CSC_index);
        free(S->columnpointer);
        free(S->rowindex);
        free(S->value_CSC);
        free(S->CSC_to_CSR_index);
        free(S->bin_rowpointer);
        free(S->bin_rowindex);

        if (S->nnzU != NULL)
        {
            free(S->nnzU);
            S->nnzU = NULL;
        }
        S->rowpointer = NULL;
        S->columnindex = NULL;
        S->value = NULL;
        S->CSR_to_CSC_index = NULL;
        S->columnpointer = NULL;
        S->rowindex = NULL;
        S->value_CSC = NULL;
        S->CSC_to_CSR_index = NULL;
        S->CUDA_rowpointer = pangulu_cuda_free_int(S->CUDA_rowpointer);
        S->CUDA_columnindex = pangulu_cuda_free_int(S->CUDA_columnindex);
        S->CUDA_columnpointer = pangulu_cuda_free_int(S->CUDA_columnpointer);
        S->CUDA_rowindex = pangulu_cuda_free_int(S->CUDA_rowindex);
        S->CUDA_CSR_to_CSC_index = pangulu_cuda_free_int(S->CUDA_CSR_to_CSC_index);
        S->CUDA_CSC_to_CSR_index = pangulu_cuda_free_int(S->CUDA_CSC_to_CSR_index);

        S->CUDA_bin_rowpointer = pangulu_cuda_free_int(S->CUDA_bin_rowpointer);
        S->CUDA_bin_rowindex = pangulu_cuda_free_int(S->CUDA_bin_rowindex);
        if (S->CUDA_nnzU != NULL)
        {
            S->CUDA_nnzU = pangulu_cuda_free_int(S->CUDA_nnzU);
        }
        S->CUDA_value = pangulu_cuda_free_double(S->CUDA_value);
        S->CUDA_value_CSC = pangulu_cuda_free_double(S->CUDA_value_CSC);
        free(S);
        return NULL;
    }
    void pangulu_destory_Big_pangulu_Smatrix(int Big_pangulu_Smatrix_length,
                                             int **Big_pangulu_rowpointer,
                                             int **Big_pangulu_columnindex,
                                             pangulu_Smatrix ***Big_pangulu_value)
    {
        int *S_Big_pangulu_rowpointer = *Big_pangulu_rowpointer;
        int *S_Big_pangulu_columnindex = *Big_pangulu_columnindex;
        pangulu_Smatrix **S_Big_pangulu_value = *Big_pangulu_value;
        for (int i = 0; i < Big_pangulu_Smatrix_length; i++)
        {
            for (int j = S_Big_pangulu_rowpointer[i]; j < S_Big_pangulu_rowpointer[i + 1]; j++)
            {
                S_Big_pangulu_value[j] = pangulu_destroy_pangulu_Smatrix(S_Big_pangulu_value[j]);
            }
        }
        free(S_Big_pangulu_rowpointer);
        free(S_Big_pangulu_columnindex);
        free(S_Big_pangulu_value);
        *Big_pangulu_rowpointer = NULL;
        *Big_pangulu_columnindex = NULL;
        *Big_pangulu_value = NULL;
    }

    pangulu_vector *pangulu_destroy_pangulu_vector(pangulu_vector *V)
    {
        free(V->value);
        V->value = NULL;
        free(V);
        return NULL;
    }

    pangulu_common *pangulu_destroy_pangulu_common(pangulu_common *common)
    {
        free(common->file_name);
        free(common);
        return NULL;
    }
#ifdef __cplusplus
}
#endif

#endif